#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from dataclasses import dataclass
from typing import List


@dataclass
class GroupValue:

    group: List[str]
    value: object

    @classmethod
    def from_json(cls, json: dict):
        if json is None:
            return None
        assert isinstance(json, dict)
        return GroupValue(
            group=json.get('group'),
            value=json.get('value')
        )

    @classmethod
    def from_json_list(cls, json_list: list):
        if json_list is None:
            return None
        assert isinstance(json_list, list)
        group_values = []
        for json in json_list:
            group_value = cls.from_json(json)
            if group_value:
                group_values.append(group_value)
        return group_values

    def to_json(self):
        return {
            'group': self.group,
            'value': self.value
        }




import json
from decimal import Decimal

from sodasql.scan.group_value import GroupValue


class JsonHelper:

    @staticmethod
    def to_json(o):
        return json.dumps(o)

    @staticmethod
    def to_json_pretty(o):
        return json.dumps(o, indent=2)

    @staticmethod
    def to_jsonnable(o):
        if o is None \
                or isinstance(o, str) \
                or isinstance(o, int) \
                or isinstance(o, float) \
                or isinstance(o, bool):
            return o
        if isinstance(o, dict):
            for key, value in o.items():
                update = False
                if not isinstance(key, str):
                    del o[key]
                    key = str(key)
                    update = True
                jsonnable_value = JsonHelper.to_jsonnable(value)
                if value is not jsonnable_value:
                    value = jsonnable_value
                    update = True
                if update:
                    o[key] = value
            return o
        if isinstance(o, list):
            for i in range(len(o)):
                element = o[i]
                jsonnable_element = JsonHelper.to_jsonnable(element)
                if element is not jsonnable_element:
                    o[i] = jsonnable_element
            return o
        if isinstance(o, Decimal):
            return float(o)
        if isinstance(o, GroupValue):
            return o.to_json()
        raise RuntimeError(f"Don't know how to jsonize {o} ({type(o)})")



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List

import pyathena

from sodasql.scan.dialect import Dialect, ATHENA, KEY_WAREHOUSE_TYPE
from sodasql.scan.parser import Parser


class AthenaDialect(Dialect):

    def __init__(self, parser: Parser):
        super().__init__(ATHENA)
        self.aws_credentials = parser.get_aws_credentials_optional()
        self.athena_staging_dir = parser.get_str_required_env('staging_dir')
        self.database = parser.get_str_required_env('database')
        self.schema = parser.get_str_required_env('schema')

    def default_connection_properties(self, params: dict):
        return {
            KEY_WAREHOUSE_TYPE: ATHENA,
            'access_key_id': 'env_var(ATHENA_ACCESS_KEY_ID)',
            'secret_access_key': 'env_var(ATHENA_SECRET_ACCESS_KEY)',
            'role_arn': 'Eg arn:aws:iam::123456789012:readonly',
            'region': 'Eg eu-west-1',
            'staging_dir': 'Eg s3://your_bucket_name/your_path',
            'database': params.get('database', 'Eg your_athena_db')
        }

    def default_env_vars(self, params: dict):
        return {
            'ATHENA_ACCESS_KEY_ID': '...',
            'ATHENA_SECRET_ACCESS_KEY': '...'
        }

    def create_connection(self):
        # pyathena.connect will do the role resolving
        # aws_credentials = self.aws_credentials.resolve_role('soda_scan')
        return pyathena.connect(
            aws_access_key_id=self.aws_credentials.access_key_id,
            aws_secret_access_key=self.aws_credentials.secret_access_key,
            s3_staging_dir=self.athena_staging_dir,
            region_name=self.aws_credentials.region_name,
            role_arn=self.aws_credentials.role_arn)

    def sql_tables_metadata_query(self, limit: str = 10, filter: str = None):
        # Alternative ( https://github.com/sodadata/soda-sql/pull/98/files )
        # return (f"SHOW tables IN `{self.database.lower()}`;")
        return (f"SELECT table_name \n"
                f"FROM information_schema.tables \n"
                f"WHERE lower(table_schema) = '{self.database.lower()}';")

    def sql_columns_metadata_query(self, table_name: str):
        return (f"SELECT column_name, data_type, is_nullable \n"
                f"FROM information_schema.columns \n"
                f"WHERE table_name = '{table_name.lower()}' \n"
                f"  AND table_schema = '{self.database.lower()}';")

    def qualify_column_name(self, column_name):
        return f'"{column_name}"'

    def qualify_table_name(self, table_name: str) -> str:
        return f'"{self.database}"."{table_name}"'

    def qualify_writable_table_name(self, table_name: str) -> str:
        return f'`{self.database}`.`{table_name}`'

    def sql_expr_avg(self, expr: str):
        return f"AVG(CAST({expr} as DECIMAL(38, 0)))"

    def sql_expr_sum(self, expr: str):
        return f"SUM(CAST ({expr} as DECIMAL(38, 0)))"



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import List, Optional

from sodasql.scan.column_metadata import ColumnMetadata
from sodasql.scan.dialect import Dialect
from sodasql.scan.metric import Metric
from sodasql.scan.missing import Missing
from sodasql.scan.scan_yml import ScanYml
from sodasql.scan.scan_yml_column import ScanYmlColumn
from sodasql.scan.validity import Validity


class ScanColumn:
    """
    Contains column information used during the scan.
    Fields are precomputed based on the scan configuration.
    """

    def __init__(self, scan, column_metadata: ColumnMetadata):
        from sodasql.scan.scan import Scan
        self.scan: Scan = scan
        self.scan_yml: ScanYml = scan.scan_yml
        self.column = column_metadata
        self.column_name = column_metadata.name
        self.scan_yml_column: ScanYmlColumn = \
            self.scan_yml.get_scan_yaml_column(self.column_name)

        dialect = self.scan.dialect
        self.qualified_column_name = dialect.qualify_column_name(self.column_name)
        self.is_text: bool = dialect.is_text(column_metadata.type)
        self.is_number: bool = dialect.is_number(column_metadata.type)

        self.missing = self.scan_yml.get_missing(self.column_name)
        self.is_missing_metric_enabled = self.scan_yml.is_any_metric_enabled(
            [Metric.MISSING_COUNT, Metric.MISSING_PERCENTAGE,
             Metric.VALUES_COUNT, Metric.VALUES_PERCENTAGE],
            self.column_name)

        self.validity = self.scan_yml.get_validity(self.column_name)
        self.is_validity_metric_enabled = self.scan_yml.is_any_metric_enabled(
            [Metric.INVALID_COUNT, Metric.INVALID_PERCENTAGE,
             Metric.VALID_COUNT, Metric.VALID_PERCENTAGE],
            self.column_name)

        self.missing_condition, self.is_default_missing_condition = \
            self.__get_missing_condition(column_metadata, self.missing, dialect)
        self.non_missing_condition: Optional[str] = f'NOT ({self.missing_condition})'
        self.valid_condition, self.is_default_valid_condition = \
            self.__get_valid_condition(column_metadata, self.validity, dialect)
        self.non_missing_and_valid_condition: Optional[str] = \
            self.__get_non_missing_and_valid_condition(self.non_missing_condition, self.valid_condition)
        self.is_default_non_missing_and_valid_condition = \
            self.is_default_missing_condition and self.is_default_valid_condition

        self.validity_format = self.scan_yml.get_validity_format(column_metadata)
        self.is_valid_enabled = \
            (self.validity is not None and self.is_validity_metric_enabled) \
            or self.scan_yml.is_any_metric_enabled([Metric.DISTINCT, Metric.UNIQUENESS], self.column_name)

        self.is_missing_enabled = self.is_valid_enabled or self.is_missing_metric_enabled

        self.is_column_numeric_text_format = \
            isinstance(self.validity_format, str) \
            and self.validity_format.startswith('number_')

        self.numeric_expr = None

        if self.is_number:
            if self.is_default_non_missing_and_valid_condition:
                self.numeric_expr = self.qualified_column_name
            else:
                self.numeric_expr = dialect.sql_expr_conditional(self.non_missing_and_valid_condition,
                                                                 self.qualified_column_name)

        elif self.is_column_numeric_text_format:
            if self.is_default_non_missing_and_valid_condition:
                self.numeric_expr = dialect.sql_expr_cast_text_to_number(self.qualified_column_name, self.validity_format)
            else:
                self.numeric_expr = dialect.sql_expr_conditional(self.non_missing_and_valid_condition,
                                                                 dialect.sql_expr_cast_text_to_number(
                                                                     self.qualified_column_name, self.validity_format))

        self.has_numeric_values = self.is_number or self.is_column_numeric_text_format
        self.mins_maxs_limit = self.scan_yml.get_mins_maxs_limit(self.column_name)

    def is_any_metric_enabled(self, metrics: List[str]):
        for metric in metrics:
            if self.is_metric_enabled(metric):
                return True
        return False

    def is_metric_enabled(self, metric: str):
        if self.scan_yml_column \
                and self.scan_yml_column \
                and metric in self.scan_yml_column.metrics:
            return True
        if self.scan_yml \
                and self.scan_yml.metrics \
                and metric in self.scan_yml.metrics:
            return True

    @classmethod
    def __get_missing_condition(cls, column_metadata: ColumnMetadata, missing: Missing, dialect: Dialect):
        qualified_column_name = dialect.qualify_column_name(column_metadata.name)
        validity_clauses = [f'{qualified_column_name} IS NULL']
        if missing:
            if missing.values:
                sql_expr_missing_values = dialect.sql_expr_list(column_metadata, missing.values)
                validity_clauses.append(f'{qualified_column_name} IN {sql_expr_missing_values}')
            if missing.format:
                format_regex = Missing.FORMATS.get(missing.format)
                qualified_regex = dialect.qualify_regex(format_regex)
                validity_clauses.append(dialect.sql_expr_regexp_like(qualified_column_name, qualified_regex))
            if missing.regex:
                qualified_regex = dialect.qualify_regex(missing.regex)
                validity_clauses.append(dialect.sql_expr_regexp_like(qualified_column_name, qualified_regex))
        return " OR ".join(validity_clauses), len(validity_clauses) == 1

    @classmethod
    def __get_valid_condition(cls, column_metadata: ColumnMetadata, validity: Validity, dialect: Dialect):
        qualified_column_name = dialect.qualify_column_name(column_metadata.name)
        if validity is None:
            return '', True
        validity_clauses = []
        if validity.format:
            format_regex = Validity.FORMATS.get(validity.format)
            qualified_regex = dialect.qualify_regex(format_regex)
            validity_clauses.append(dialect.sql_expr_regexp_like(qualified_column_name, qualified_regex))
        if validity.regex:
            qualified_regex = dialect.qualify_regex(validity.regex)
            validity_clauses.append(dialect.sql_expr_regexp_like(qualified_column_name, qualified_regex))
        if validity.values:
            valid_values_sql = dialect.literal_list(validity.values)
            validity_clauses.append(dialect.sql_expr_in(qualified_column_name, valid_values_sql))
        if validity.min_length:
            validity_clauses.append(f'{dialect.sql_expr_length(qualified_column_name)} >= {validity.min_length}')
        if validity.max_length:
            validity_clauses.append(f'{dialect.sql_expr_length(qualified_column_name)} <= {validity.max_length}')
        if validity.min:
            validity_clauses.append(f'{qualified_column_name} >= {validity.min}')
        if validity.max:
            validity_clauses.append(f'{qualified_column_name} <= {validity.max}')
        return '(' + ' AND '.join(validity_clauses) + ')', len(validity_clauses) == 0

    def __get_non_missing_and_valid_condition(self, non_missing_condition, valid_condition):
        if non_missing_condition and valid_condition:
            return f'{non_missing_condition} AND {valid_condition}'
        elif non_missing_condition:
            return non_missing_condition
        elif valid_condition:
            return valid_condition
        return ''

    def get_group_by_cte(self):
        if self.non_missing_and_valid_condition and self.scan.table_sample_clause:
            where = f'{self.non_missing_and_valid_condition}\n    {self.scan.table_sample_clause}'
        elif self.non_missing_and_valid_condition:
            where = self.non_missing_and_valid_condition
        elif self.scan.table_sample_clause:
            where = self.scan.table_sample_clause
        else:
            where = f'{self.qualified_column_name} IS NOT NULL'

        if self.scan.filter_sql:
            where = f'{where}\n  AND {self.scan.filter_sql}'

        return (
            f"WITH group_by_value AS ( \n"
            f"  SELECT \n"
            f"    {self.qualified_column_name} AS value, \n"
            f"    COUNT(*) AS frequency \n"
            f"  FROM {self.scan.qualified_table_name} \n"
            f"  WHERE {where} \n"
            f"  GROUP BY {self.qualified_column_name} \n"
            f")"
        )

    def get_group_by_cte_numeric_value_expression(self):
        if self.is_number:
            return 'value'
        if self.is_column_numeric_text_format:
            return self.scan.dialect.sql_expr_cast_text_to_number('value', self.validity_format)

    def get_tests(self):
        return self.scan_yml_column.tests if self.scan_yml_column and self.scan_yml_column.tests else []

    def get_histogram_buckets(self) -> int:
        # TODO make configurable
        return 20

    def get_metric_value(self, metric: str):
        return self.scan.scan_result.get(metric, self.column_name)



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from dataclasses import dataclass
from typing import Optional, Any

from sodasql.scan.test import Test


@dataclass
class TestResult:

    test: Test
    passed: bool
    value: Optional[Any] = None
    error: Optional[str] = None
    group_values: Optional[dict] = None

    def to_json(self):
        if not self.test or not self.test.expression:
            return {
                'error': 'Invalid test result'
            }
        if self.error:
            return {
                'test': self.test.expression,
                'error': error
            }
        test_result_json = {
            'test': self.test.expression,
            'passed': self.passed,
            'value': self.value
        }
        if self.group_values:
            test_result_json['groupValues'] = self.group_values
        return test_result_json



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
import os
from typing import List

from sodasql.common.yaml_helper import YamlHelper
from sodasql.scan.file_system import FileSystemSingleton
from sodasql.scan.parser import Parser
from sodasql.scan.scan import Scan
from sodasql.scan.scan_yml import ScanYml
from sodasql.scan.warehouse_yml import WarehouseYml
from sodasql.scan.warehouse_yml_parser import read_warehouse_yml_file
from sodasql.soda_server_client.soda_server_client import SodaServerClient


class ScanBuilder:
    """
    Programmatic scan execution based on default dir structure:

    scan_builder = ScanBuilder()
    scan_builder.scan_yml_file = 'tables/my_table.yml'
    # scan_builder will automatically find the warehouse.yml in the parent and same directory as the scan YAML file
    # scan_builder.warehouse_yml_file = '../warehouse.yml'
    scan = scan_builder.build()
    scan_result = scan.execute()
    if scan_result.has_failures():
        print('Scan has test failures, stop the pipeline')

    Programmatic scan execution using dicts:

    scan_builder = ScanBuilder()
    scan_builder.warehouse_dict = {
        'name': 'my_warehouse_name',
        'connection': {
            'type': 'snowflake',
            ...
        }
    })
    scan_builder.scan_dict = {
        ...
    }
    scan = scan_builder.build()
    scan_result = scan.execute()
    if scan_result.has_failures():
        print('Scan has test failures, stop the pipeline')
    """

    def __init__(self):
        self.file_system = FileSystemSingleton.INSTANCE
        self.warehouse_yml_file: str = None
        self.warehouse_yml_dict: dict = None
        self.warehouse_yml: WarehouseYml = None
        self.scan_yml_file: str = None
        self.time: str = None
        self.scan_yml_dict: dict = None
        self.scan_yml: ScanYml = None
        self.variables: dict = {}
        self.parsers: List[Parser] = []
        self.assert_no_warnings_or_errors = True
        self.soda_server_client: SodaServerClient = None

    def build(self):
        self._build_warehouse_yml()
        self._build_scan_yml()

        for parser in self.parsers:
            parser.assert_no_warnings_or_errors()
        if not self.scan_yml or not self.warehouse_yml:
            return

        from sodasql.scan.warehouse import Warehouse
        warehouse = Warehouse(self.warehouse_yml)

        self._create_soda_server_client()
        
        return Scan(warehouse=warehouse,
                    scan_yml=self.scan_yml,
                    variables=self.variables,
                    soda_server_client=self.soda_server_client,
                    time=self.time)

    def _build_warehouse_yml(self):
        if not self.warehouse_yml_file and not self.warehouse_yml_dict and not self.warehouse_yml:
            logging.error(f'No warehouse specified')
            return

        elif self.warehouse_yml_file and not self.warehouse_yml_dict and not self.warehouse_yml:
            if not isinstance(self.warehouse_yml_file, str):
                logging.error(f'scan_builder.warehouse_yml_file must be str, but was {type(self.warehouse_yml_file)}: {self.warehouse_yml_file}')
            else:
                self.warehouse_yml_dict = read_warehouse_yml_file(self.warehouse_yml_file)

        if self.warehouse_yml_dict and not self.warehouse_yml:
            from sodasql.scan.warehouse_yml_parser import WarehouseYmlParser
            warehouse_parser = WarehouseYmlParser(self.warehouse_yml_dict, self.warehouse_yml_file)
            warehouse_parser.log()
            self.parsers.append(warehouse_parser)
            self.warehouse_yml = warehouse_parser.warehouse_yml

    def _build_scan_yml(self):
        file_system = FileSystemSingleton.INSTANCE

        if not self.scan_yml_file and not self.scan_yml_dict and not self.scan_yml:
            logging.error(f'No scan specified')
            return

        elif self.scan_yml_file and not self.scan_yml_dict and not self.scan_yml:
            if not isinstance(self.scan_yml_file, str):
                logging.error(f'scan_builder.scan_yml_file must be str, but was {type(self.scan_yml_file)}: {self.scan_yml_file}')
            elif file_system.is_readable_file(self.scan_yml_file):
                scan_yml_str = self.file_system.file_read_as_str(self.scan_yml_file)
                if scan_yml_str:
                    self.scan_yml_dict = YamlHelper.parse_yaml(scan_yml_str, self.scan_yml_file)
                else:
                    logging.error(f'Failed to file scan yaml file: {self.scan_yml_file}')

        if self.scan_yml_dict and not self.scan_yml:
            from sodasql.scan.scan_yml_parser import ScanYmlParser
            scan_yml_parser = ScanYmlParser(self.scan_yml_dict, self.scan_yml_file)
            scan_yml_parser.log()
            self.parsers.append(scan_yml_parser)
            self.scan_yml = scan_yml_parser.scan_yml

    def _create_soda_server_client(self):
        if not self.soda_server_client:
            if self.warehouse_yml.soda_api_key_id and self.warehouse_yml.soda_api_key_secret:
                host = self.warehouse_yml.soda_host
                api_key_id = self.warehouse_yml.soda_api_key_id
                api_key_secret = self.warehouse_yml.soda_api_key_secret
                port = str(self.warehouse_yml.soda_port)
                protocol = self.warehouse_yml.soda_protocol
            else:
                host = os.getenv('SODA_HOST', 'cloud.soda.io')
                api_key_id = os.getenv('SODA_SERVER_API_KEY_ID', None)
                api_key_secret = os.getenv('SODA_SERVER_API_KEY_SECRET', None)
                port = os.getenv('SODA_PORT', '443')
                protocol = os.getenv('SODA_PROTOCOL', 'https')

            if api_key_id and api_key_secret:
                self.soda_server_client = SodaServerClient(host,
                                                           api_key_id=api_key_id,
                                                           api_key_secret=api_key_secret,
                                                           protocol=protocol,
                                                           port=port)
            else:
                logging.debug("No Soda Cloud account configured")



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
import re
from numbers import Number
from typing import List

from sodasql.scan.column_metadata import ColumnMetadata
from sodasql.scan.parser import Parser
from sodasql.scan.scan_yml import ScanYml

KEY_WAREHOUSE_TYPE = 'type'

POSTGRES = 'postgres'
SNOWFLAKE = 'snowflake'
REDSHIFT = 'redshift'
BIGQUERY = 'bigquery'
ATHENA = 'athena'

ALL_WAREHOUSE_TYPES = [POSTGRES,
                       SNOWFLAKE,
                       REDSHIFT,
                       BIGQUERY,
                       ATHENA]


class Dialect:

    data_type_varchar_255 = "VARCHAR(255)"
    data_type_integer = "INTEGER"
    data_type_bigint = "BIGINT"
    data_type_decimal = "REAL"

    def __init__(self, type: str):
        self.type = type

    @classmethod
    def create(cls, parser: Parser):
        warehouse_type = parser.get_str_optional(KEY_WAREHOUSE_TYPE)
        if warehouse_type == POSTGRES:
            from sodasql.dialects.postgres_dialect import PostgresDialect
            return PostgresDialect(parser)
        if warehouse_type == SNOWFLAKE:
            from sodasql.dialects.snowflake_dialect import SnowflakeDialect
            return SnowflakeDialect(parser)
        if warehouse_type == REDSHIFT:
            from sodasql.dialects.redshift_dialect import RedshiftDialect
            return RedshiftDialect(parser)
        if warehouse_type == BIGQUERY:
            from sodasql.dialects.bigquery_dialect import BigQueryDialect
            return BigQueryDialect(parser)
        if warehouse_type == ATHENA:
            from sodasql.dialects.athena_dialect import AthenaDialect
            return AthenaDialect(parser)

    @classmethod
    def create_for_warehouse_type(cls, warehouse_type):
        from sodasql.scan.dialect_parser import DialectParser
        return cls.create(DialectParser(warehouse_connection_dict={KEY_WAREHOUSE_TYPE: warehouse_type}))

    def default_connection_properties(self, params: dict):
        pass

    def default_env_vars(self, params: dict):
        pass

    # TODO
    def sql_connection_test(self):
        pass

    def create_connection(self):
        raise RuntimeError('TODO override and implement this abstract method')

    def create_scan(self, *args, **kwargs):
        # Purpose of this method is to enable dialects to override and customize the scan implementation
        from sodasql.scan.scan import Scan
        return Scan(*args, **kwargs)

    def is_text(self, column_type: str):
        for text_type in self._get_text_types():
            if column_type and text_type.upper() in column_type.upper():
                return True
        return False

    def _get_text_types(self):
        return ['CHAR', 'TEXT', 'STRING']

    def is_number(self, column_type: str):
        for number_type in self._get_number_types():
            if column_type and number_type.upper() in column_type.upper():
                return True
        return False

    def _get_number_types(self):
        return ['INT', 'REAL', 'PRECISION', 'NUMBER', 'DECIMAL']

    def sql_columns_metadata_query(self, table_name: str) -> str:
        raise RuntimeError('TODO override and implement this abstract method')

    def sql_tables_metadata_query(self, limit: str = 10, filter: str = None):
        raise RuntimeError('TODO override and implement this abstract method')

    def sql_expr_count_all(self) -> str:
        return 'COUNT(*)'

    def sql_expr_count_conditional(self, condition: str):
        return f'COUNT(CASE WHEN {condition} THEN 1 END)'

    def sql_expr_conditional(self, condition: str, expr: str):
        return f'CASE WHEN {condition} THEN {expr} END'

    def sql_expr_count(self, expr):
        return f'COUNT({expr})'

    def sql_expr_distinct(self, expr):
        return f'DISTINCT({expr})'

    def sql_expr_min(self, expr):
        return f'MIN({expr})'

    def sql_expr_length(self, expr):
        return f'LENGTH({expr})'

    def sql_expr_min(self, expr: str):
        return f'MIN({expr})'

    def sql_expr_max(self, expr: str):
        return f'MAX({expr})'

    def sql_expr_avg(self, expr: str):
        return f'AVG({expr})'

    def sql_expr_sum(self, expr: str):
        return f'SUM({expr})'

    def sql_expr_variance(self, expr: str):
        return f'VARIANCE({expr})'

    def sql_expr_stddev(self, expr: str):
        return f'STDDEV({expr})'

    def sql_expr_regexp_like(self, expr: str, pattern: str):
        return f"REGEXP_LIKE({expr}, '{self.qualify_regex(pattern)}')"

    def sql_expr_list(self, column: ColumnMetadata, values: List[str]) -> str:
        if self.is_text(column.type):
            sql_values = [self.literal_string(value) for value in values]
        elif self.is_number(column.type):
            sql_values = [self.literal_number(value) for value in values]
        else:
            raise RuntimeError(f"Couldn't format list {str(values)} for column {str(column)}")
        return '('+','.join(sql_values)+')'

    def sql_expr_cast_text_to_number(self, quoted_column_name, validity_format):
        if validity_format == 'number_whole':
            return f"CAST({quoted_column_name} AS {self.data_type_decimal})"
        not_number_pattern = self.qualify_regex(r"[^-\d\.\,]")
        comma_pattern = self.qualify_regex(r"\,")
        return f"CAST(REGEXP_REPLACE(REGEXP_REPLACE({quoted_column_name}, '{not_number_pattern}', ''), "\
               f"'{comma_pattern}', '.') AS {self.data_type_decimal})"

    def literal_number(self, value: Number):
        if value is None:
            return None
        return str(value)

    def literal_string(self, value: str):
        if value is None:
            return None
        return "'"+self.escape_metacharacters(value)+"'"

    def literal_list(self, l: list):
        if l is None:
            return None
        return '(' + (','.join([self.literal(e) for e in l])) + ')'

    def literal(self, o: object):
        if isinstance(o, Number):
            return self.literal_number(o)
        elif isinstance(o, str):
            return self.literal_string(o)
        elif isinstance(o, list) or isinstance(o, set) or isinstance(o, tuple):
            return self.literal_list(o)
        raise RuntimeError(f'Cannot convert type {type(o)} to a SQL literal: {o}')

    def qualify_table_name(self, table_name: str) -> str:
        return table_name

    def qualify_writable_table_name(self, table_name: str) -> str:
        return table_name

    def qualify_regex(self, regex):
        return regex

    def qualify_string(self, value: str):
        return value

    def qualify_column_name(self, column_name: str):
        return column_name

    @staticmethod
    def escape_metacharacters(value: str):
        return re.sub(r'(\\.)', r'\\\1', value)

    def sql_declare_string_column(self, column_name):
        return f"{column_name} {self.data_type_varchar_255}"

    def sql_declare_integer_column(self, column_name):
        return f"{column_name} {self.data_type_integer}"

    def sql_declare_decimal_column(self, column_name):
        return f"{column_name} {self.data_type_decimal}"

    def sql_declare_big_integer_column(self, column_name):
        return f"{column_name} {self.data_type_bigint}"

    def sql_expression(self, expression_dict: dict):
        if expression_dict is None:
            return None
        type = expression_dict['type']
        if type == 'number':
            sql = self.literal_number(expression_dict['value'])
        elif type == 'string':
            sql = self.literal_string(expression_dict['value'])
        elif type == 'columnValue':
            sql = expression_dict['columnName']
        elif type == 'collection':
            # collection of string or number literals
            value = expression_dict['value']
            sql = self.literal_list(value)
        elif type == 'equals':
            left = self.sql_expression(expression_dict['left'])
            right = self.sql_expression(expression_dict['right'])
            sql = self.sql_expr_equal(left, right)
        elif type == 'lessThan':
            left = self.sql_expression(expression_dict['left'])
            right = self.sql_expression(expression_dict['right'])
            sql = self.sql_expr_less_than(left, right)
        elif type == 'lessThanOrEqual':
            left = self.sql_expression(expression_dict['left'])
            right = self.sql_expression(expression_dict['right'])
            sql = self.sql_expr_less_than_or_equal(left, right)
        elif type == 'greaterThan':
            left = self.sql_expression(expression_dict['left'])
            right = self.sql_expression(expression_dict['right'])
            sql = self.sql_expr_greater_than(left, right)
        elif type == 'greaterThanOrEqual':
            left = self.sql_expression(expression_dict['left'])
            right = self.sql_expression(expression_dict['right'])
            sql = self.sql_expr_greater_than_or_equal(left, right)
        elif type == 'between':
            clauses = []
            value = self.sql_expression(expression_dict['value'])
            gte = self.literal_number(expression_dict.get('gte'))
            gt = self.literal_number(expression_dict.get('gt'))
            lte = self.literal_number(expression_dict.get('lte'))
            lt = self.literal_number(expression_dict.get('lt'))
            if gte:
                clauses.append(self.sql_expr_less_than_or_equal(gte, value))
            elif gt:
                clauses.append(self.sql_expr_less_than(gt, value))
            if lte:
                clauses.append(self.sql_expr_less_than_or_equal(value, lte))
            elif lt:
                clauses.append(self.sql_expr_less_than(value, lt))
            sql = ' AND '.join(clauses)
        elif type == 'in':
            left = self.sql_expression(expression_dict['left'])
            right = self.sql_expression(expression_dict['right'])
            sql = self.sql_expr_in(left, right)
        elif type == 'contains':
            value = self.sql_expression(expression_dict['left'])
            substring = self.escape_metacharacters(expression_dict['right']['value'])
            sql = self.sql_expr_contains(value, substring)
        elif type == 'startsWith':
            value = self.sql_expression(expression_dict['left'])
            substring = self.escape_metacharacters(expression_dict['right']['value'])
            sql = self.sql_expr_starts_with(value, substring)
        elif type == 'endsWith':
            value = self.sql_expression(expression_dict['left'])
            substring = self.escape_metacharacters(expression_dict['right']['value'])
            sql = self.sql_expr_ends_with(value, substring)
        elif type == 'not':
            sql = 'NOT (' + self.sql_expression(expression_dict['expression']) + ')'
        elif type == 'and':
            sql = '(' + (') AND ('.join([self.sql_expression(e) for e in expression_dict['andExpressions']])) + ')'
        elif type == 'or':
            sql = '(' + (') OR ('.join([self.sql_expression(e) for e in expression_dict['orExpressions']])) + ')'
        else:
            raise RuntimeError(f'Unsupported expression type: {type}')
        return sql

    def sql_expr_equal(self, left, right):
        return f'{left} = {right}'

    def sql_expr_less_than(self, left, right):
        return f'{left} < {right}'

    def sql_expr_less_than_or_equal(self, left, right):
        return f'{left} <= {right}'

    def sql_expr_greater_than(self, left, right):
        return f'{left} > {right}'

    def sql_expr_greater_than_or_equal(self, left, right):
        return f'{left} >= {right}'

    def sql_expr_in(self, left, right):
        return f'{left} IN {right}'

    def sql_expr_contains(self, value, substring):
        return value + " LIKE '%" + substring + "%'"

    def sql_expr_starts_with(self, value, substring):
        return value + " LIKE '" + substring + "%'"

    def sql_expr_ends_with(self, value, substring):
        return value + " LIKE '%" + substring + "'"

    def get_type_name(self, column_description):
        return str(column_description[1])



import logging

import yaml


class YamlHelper:

    @staticmethod
    def parse_yaml(yaml_str: str, description: str = None):
        try:
            return yaml.load(yaml_str, Loader=yaml.FullLoader)
        except Exception as e:
            logging.error(f'Parsing YAML failed: {str(e)}: ({description if description else yaml_str})xWW4')



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import json
import logging
import os
import traceback
from collections import deque
from dataclasses import dataclass
from typing import List, Deque, Optional

import yaml

from sodasql.scan.aws_credentials import AwsCredentials
from sodasql.scan.metric import Metric
from sodasql.scan.test import Test

ERROR = 'error'
WARNING = 'warning'
INFO = 'info'


class ParseLog:
    def __init__(self, level: str, message: str):
        self.level = level
        self.message = message

    def __str__(self):
        return f'[{self.level}] {self.message}'

    def log(self):
        if self.level == ERROR:
            logging.error(self.message)
        elif self.level == WARNING:
            logging.warning(self.message)
        elif self.level == INFO:
            logging.info(self.message)
        else:
            logging.debug(self.message)


@dataclass
class ParseContext:
    name: str
    object: dict


class Parser:

    def __init__(self, description: str):
        self.description: str = description
        self.contexts: Deque[ParseContext] = deque()
        self.logs: List[ParseLog] = []

    def __str__(self):
        return '\n'.join([str(log) for log in self.logs])

    def _read_file_as_string(self, file_path: str):
        try:
            with open(file_path) as f:
                return f.read()
        except Exception as e:
            self.error(f"Couldn't read file {file_path}: {str(e)}")

    def _parse_yaml_str(self, yaml_str):
        try:
            return yaml.load(yaml_str, Loader=yaml.FullLoader)
        except Exception as e:
            self.error(f"Couldn't parse yaml in {self.description}: {str(e)}")

    def _push_context(self, object=None, name: Optional[object] = None):
        name = name if isinstance(name, str) or name is None else str(name)
        self.contexts.append(ParseContext(object=object, name=name))

    def _pop_context(self):
        return self.contexts.pop()

    def _get_context_description(self):
        return '.'.join([context.name for context in self.contexts if context.name])

    def error(self, message: str):
        return self.logs.append(ParseLog(ERROR, message))

    def warning(self, message: str):
        return self.logs.append(ParseLog(WARNING, message))

    def info(self, message):
        return self.logs.append(ParseLog(INFO, message))

    def log(self):
        for log in self.logs:
            log.log()

    def check_invalid_keys(self, valid_keys: List[str]):
        """
        Adds a warning for all invalid configured property names
        """
        context_iterable = self._get_current_context_object()
        for invalid_key in [configured_key for configured_key in context_iterable if configured_key not in valid_keys]:
            self.warning(f'Invalid key in {self._get_context_description()} : {invalid_key}')

    def has_warnings_or_errors(self):
        for log in self.logs:
            if log.level == ERROR or log.level == WARNING:
                return True
        return False

    def assert_no_warnings_or_errors(self):
        if self.has_warnings_or_errors():
            raise AssertionError(f'{self.description} configuration errors: \n  '
                                 + ('\n  '.join([str(log) for log in self.logs])))

    def get_str_required(self, property_name: str):
        return self._get(property_name, str, True)

    def get_str_optional(self, property_name: str, default=None):
        return self._get(property_name, str, False, default)

    def get_str_required_env(self, property_name: str):
        return self._get(property_name, str, True, resolve_env=True)

    def get_credential(self, property_name: str):
        return self._get(property_name, str, False, resolve_env=True, env_required=True)

    def get_str_optional_env(self, property_name: str, default=None):
        return self._get(property_name, str, False, default, resolve_env=True)

    def get_int_required(self, property_name: str):
        return self._get(property_name, int, True)

    def get_int_optional(self, property_name: str, default=None):
        return self._get(property_name, int, False, default)

    def get_float_required(self, property_name: str):
        return self._get(property_name, float, True)

    def get_float_optional(self, property_name: str, default=None):
        return self._get(property_name, float, False, default)

    def get_dict_required(self, property_name: str):
        return self._get(property_name, dict, True)

    def get_dict_optional(self, property_name: str, default=None):
        return self._get(property_name, dict, False, default)

    def get_list_required(self, property_name: str):
        return self._get(property_name, list, True)

    def get_list_optional(self, property_name: str, default=None):
        return self._get(property_name, list, False, default)

    def get_aws_credentials_optional(self):
        access_key_id = self.get_str_optional_env('access_key_id')
        if access_key_id:
            return AwsCredentials(
                access_key_id=access_key_id,
                secret_access_key=self.get_credential('secret_access_key'),
                role_arn=self.get_str_optional_env('role_arn'),
                session_token=self.get_credential('session_token'),
                region_name=self.get_str_optional_env('region', 'eu-west-1'))

    def get_file_json_dict_required(self, property_name: str):
        file_str: str = self._read_file_as_string(property_name)
        try:
            return json.loads(file_str)
        except Exception as e:
            self.error(f"Couldn't parse json configuration {property_name} for {self.description}: {str(e)}")

    def _get(self,
             property_name: str,
             return_type: type,
             is_required: bool,
             default=None,
             resolve_env: bool = False,
             env_required: bool = False):
        properties = self._get_current_context_object()
        if property_name in properties:
            value = properties.get(property_name)

            if resolve_env:
                if isinstance(value, str) \
                        and value.strip().startswith('env_var(') \
                        and value.strip().endswith(')'):
                    env_var_name = value.strip()[len('env_var('):-1]
                    value = os.getenv(env_var_name)
                    if not value:
                        self.info(f'Environment variable {env_var_name} is not set')
                elif env_required:
                    self.error(f'{self._get_context_description()}.{property_name} is considered a credential and '
                               f'must be passed as an environment variable eg: env_var(YOUR_ENVIRONMENT_VAR)')

            if value is None or isinstance(value, return_type):
                return value
            try:
                if return_type == int:
                    return int(value)
                if return_type == str:
                    return str(value)
                if return_type == float:
                    return float(value)
                raise ValueError(str(type(value)))
            except ValueError as e:
                self.error(f'Invalid {property_name}: Expected {str(return_type)}, but was {str(value)}: {str(e)}')
                return default

        else:
            if is_required:
                self.error(f'Property {property_name} does not exist in {self._get_context_description()}')
            return default

    def _get_current_context_object(self):
        return self.contexts[-1].object

    def parse_tests(self,
                    parent_dict: dict,
                    tests_key: str,
                    context_table_name: Optional[str] = None,
                    context_column_name: Optional[str] = None,
                    context_sql_metric: Optional[str] = None) -> List[Test]:
        tests: List[Test] = []

        test_ymls = parent_dict.get(tests_key)

        self._push_context(None, tests_key)
        try:
            if isinstance(test_ymls, list):
                for test_index in range(len(test_ymls)):
                    test_yml = test_ymls[test_index]
                    self._push_context(None, str(test_index))
                    try:
                        test_name = f'test_{test_index}'
                        test_expression = test_yml
                        test = self.parse_test(test_name,
                                               test_expression,
                                               context_table_name,
                                               context_column_name,
                                               context_sql_metric)
                        if test:
                            tests.append(test)
                    finally:
                        self._pop_context()

            elif isinstance(test_ymls, dict):
                for test_name in test_ymls:
                    self._push_context(None, test_name)
                    try:
                        test_expression = test_ymls[test_name]
                        test = self.parse_test(test_name,
                                               test_expression,
                                               context_table_name,
                                               context_column_name,
                                               context_sql_metric)
                        if test:
                            tests.append(test)
                    finally:
                        self._pop_context()

            elif test_ymls is not None:
                self.error(f'Tests should be either a list of test expressions or an object of named test expressions: {test_ymls} ({str(type(test_ymls))})')
        finally:
            self._pop_context()

        return tests

    def parse_test(self,
                   test_name: str,
                   test_expression: str,
                   context_table_name: str = None,
                   context_column_name: str = None,
                   context_sql_metric: str = None):

        if not test_name:
            self.error('Test name is required')
            return

        if not test_expression:
            self.error('Test expression is required')
            return

        test_description = self.create_test_description(
            test_expression,
            context_table_name,
            context_column_name,
            context_sql_metric)

        try:
            compiled_code = compile(test_expression, 'test', 'eval')
            first_metric = None

            if context_table_name or context_column_name:
                names = compiled_code.co_names
                first_metric = names[0]
                non_metric_names = [name for name in names if name not in Metric.METRIC_TYPES]
                if len(non_metric_names) != 0 or len(names) == 0:
                    # Dunno yet if this should be info, warning or error.  So for now keeping it open.
                    # SQL metric names and variables are not known until eval.
                    self.info(f'At least one of the variables used in test ({set(names)}) '
                              f'was not a valid metric type. Metric types: {Metric.METRIC_TYPES}, '
                              f'Test: {test_description}')

            return Test(test_description,
                        test_expression,
                        first_metric,
                        context_column_name)

        except SyntaxError:
            stacktrace_lines = traceback.format_exc().splitlines()
            self.error(f'Syntax error in test {test_description}:\n' +
                       ('\n'.join(stacktrace_lines[-3:])))

    def create_test_description(self,
                                test_expression,
                                context_table_name: str,
                                context_column_name: str,
                                context_sql_metric: str):
        parts = []
        if context_table_name:
            parts.append(f'table({context_table_name})')
        if context_column_name:
            parts.append(f'column({context_column_name})')
        if context_sql_metric:
            parts.append(f'sql_metric({context_sql_metric})')
        parts.append(f'expression({test_expression})')
        return ' '.join(parts)



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import sys


class LoggingHelper:

    @classmethod
    def configure_for_cli(cls):
        cls.configure_for_test()

    @classmethod
    def configure_for_test(cls):
        logging.getLogger('urllib3').setLevel(logging.WARNING)
        logging.getLogger('botocore').setLevel(logging.WARNING)
        logging.getLogger('pyathena').setLevel(logging.WARNING)
        logging.getLogger('faker').setLevel(logging.ERROR)
        logging.getLogger('snowflake').setLevel(logging.WARNING)
        logging.basicConfig(
            level=logging.DEBUG,
            # https://docs.python.org/3/library/logging.html#logrecord-attributes
            # %(name)s
            format="  | %(message)s",
            handlers=[logging.StreamHandler(sys.stdout)]
        )



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import List, Optional


class Missing:

    FORMATS = {
        # more regexes: https://regexr.com/ https://digitalfortress.tech/tricks/top-15-commonly-used-regex/ http://regexlib.com/
        'empty': r'^$',
        'whitespace': r'^\s*$'
    }

    def __init__(self):
        self.values: Optional[List[str]] = None
        self.format: Optional[str] = None
        self.regex: Optional[str] = None



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Optional

import boto3


class AwsCredentials:

    def __init__(self,
                 access_key_id: Optional[str] = None,
                 secret_access_key: Optional[str] = None,
                 role_arn: Optional[str] = None,
                 session_token: Optional[str] = None,
                 region_name: Optional[str] = 'eu-west-1'):
        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key
        self.role_arn = role_arn
        self.session_token = session_token
        self.region_name = region_name

    @classmethod
    def from_configuration(cls, configuration: dict):
        """
        An AwsCredentials if there is an access_key_id specified, None otherwise
        """
        access_key_id = configuration.get('access_key_id')
        if not isinstance(access_key_id, str):
            return None
        return AwsCredentials(
            access_key_id=access_key_id,
            secret_access_key=configuration.get('secret_access_key'),
            role_arn=configuration.get('role_arn'),
            session_token=configuration.get('session_token'),
            region_name=configuration.get('region', 'eu-west-1'))

    def resolve_role(self, role_session_name: str):
        if self.has_role():
            return self.assume_role(role_session_name)
        return self

    def has_role(self):
        return isinstance(self.role_arn, str)

    def assume_role(self, role_session_name: str):
        self.region_name = self.region_name
        self.sts_client = boto3.client(
            'sts',
            region_name=self.region_name,
            aws_access_key_id=self.access_key_id,
            aws_secret_access_key=self.secret_access_key,
            aws_session_token=self.session_token)

        assumed_role_object = self.sts_client.assume_role(
            RoleArn=self.role_arn,
            RoleSessionName=role_session_name
        )
        credentials_dict = assumed_role_object['Credentials']
        return AwsCredentials(
            region_name=self.region_name,
            access_key_id=credentials_dict['AccessKeyId'],
            secret_access_key=credentials_dict['SecretAccessKey'],
            session_token=credentials_dict['SessionToken'])



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from snowflake import connector

from sodasql.scan.dialect import Dialect, SNOWFLAKE, KEY_WAREHOUSE_TYPE
from sodasql.scan.parser import Parser
from sodasql.scan.scan_yml import ScanYml


class SnowflakeDialect(Dialect):

    def __init__(self, parser: Parser):
        super().__init__(SNOWFLAKE)
        self.account = parser.get_str_required_env('account')
        self.warehouse = parser.get_str_required_env('warehouse')
        self.username = parser.get_str_required_env('username')
        self.password = parser.get_credential('password')
        self.database = parser.get_str_optional_env('database')
        self.schema = parser.get_str_required_env('schema')

    def default_connection_properties(self, params: dict):
        return {
            KEY_WAREHOUSE_TYPE: SNOWFLAKE,
            'username': 'env_var(SNOWFLAKE_USERNAME)',
            'password': 'env_var(SNOWFLAKE_PASSWORD)',
            'account': 'YOURACCOUNT.eu-central-1',
            'database': params.get('database','YOUR_DATABASE'),
            'warehouse': 'YOUR_WAREHOUSE',
            'schema': 'PUBLIC'
        }

    def default_env_vars(self, params: dict):
        return {
            'SNOWFLAKE_USERNAME': params.get('username', 'YOUR_SNOWFLAKE_USERNAME_GOES_HERE'),
            'SNOWFLAKE_PASSWORD': params.get('password', 'YOUR_SNOWFLAKE_PASSWORD_GOES_HERE')
        }

    def create_connection(self):
        return connector.connect(
            user=self.username,
            password=self.password,
            account=self.account,
            warehouse=self.warehouse,
            database=self.database,
            schema=self.schema
        )

    def sql_tables_metadata_query(self, limit: str = 10, filter: str = None):
        sql = (f"SELECT table_name \n" 
               f"FROM information_schema.tables \n" 
               f"WHERE lower(table_schema)='{self.schema.lower()}'")
        if self.database:
            sql += f" \n  AND lower(table_catalog) = '{self.database.lower()}'"
        if isinstance(limit, int):
            sql += f" \nLIMIT {limit}"
        return sql

    def sql_columns_metadata_query(self, table_name: str) -> str:
        sql = (f"SELECT column_name, data_type, is_nullable \n"
               f'FROM information_schema.columns \n'
               f"WHERE lower(table_name) = '{table_name.lower()}'")
        if self.database:
            sql += f" \n  AND lower(table_catalog) = '{self.database.lower()}'"
        if self.schema:
            sql += f" \n  AND lower(table_schema) = '{self.schema.lower()}'"
        return sql

    def qualify_regex(self, regex) -> str:
        return self.escape_metacharacters(regex)

    def qualify_column_name(self, column_name: str):
        return f'"{column_name}"'

    def qualify_table_name(self, table_name: str) -> str:
        return f'"{table_name.upper()}"'



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
import os
from pathlib import Path


class FileSystem:

    def join(self, a: str, *paths: str):
        return os.path.join(a, *paths)

    def dirname(self, path: str):
        return os.path.dirname(path)

    def normpath(self, path):
        return os.path.normpath(path)

    def split(self, path: str):
        return os.path.split(path)

    def file_exists(self, path: str):
        expanded_path = os.path.expanduser(path)
        return Path(expanded_path).exists()

    def is_dir(self, path: str):
        expanded_path = os.path.expanduser(path)
        return Path(expanded_path).is_dir()

    def is_file(self, path):
        expanded_path = os.path.expanduser(path)
        return Path(expanded_path).is_file()

    def is_readable(self, path: str):
        return os.access(path, os.R_OK)

    def list_dir(self, dir_path):
        return os.listdir(dir_path)

    def mkdirs(self, path: str):
        expanded_path = os.path.expanduser(path)
        Path(expanded_path).mkdir(parents=True, exist_ok=True)

    def user_home_dir(self):
        return str(Path.home())

    def file_read_as_str(self, path: str) -> str:
        expanded_path = os.path.expanduser(path)
        try:
            with open(expanded_path) as f:
                return f.read()
        except Exception as e:
            logging.debug(f"Couldn't read {str(path)}: {str(e)}")

    def file_write_from_str(self, path: str, file_content_str):
        expanded_path = os.path.expanduser(path)
        path_path: Path = Path(expanded_path)
        is_new = not path_path.exists()
        try:
            with open(path_path, 'w+') as f:
                f.write(file_content_str)
            if is_new:
                os.chmod(path, 0o666)
        except Exception as e:
            logging.debug(f"Couldn't write {str(path)}: {str(e)}")

    def is_readable_file(self, file_path: str):
        if not self.file_exists(file_path):
            logging.error(f'{file_path} does not exist')
            return False
        elif not self.is_file(file_path):
            logging.error(f'{file_path} is not a file')
            return False
        elif not self.is_readable(file_path):
            logging.error(f'{file_path} is not readable')
            return False
        return True


class FileSystemSingleton:
    INSTANCE: FileSystem = FileSystem()


def file_system():
    return FileSystemSingleton.INSTANCE



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from sodasql.scan.file_system import FileSystemSingleton
from sodasql.scan.parser import Parser
from sodasql.scan.sql_metric_yml import SqlMetricYml

KEY_SQL = 'sql'
KEY_METRIC_NAMES = 'metric_names'
KEY_TESTS = 'tests'
KEY_GROUP_FIELDS = 'group_fields'

VALID_SQL_METRIC_KEYS = [KEY_SQL, KEY_METRIC_NAMES, KEY_TESTS, KEY_GROUP_FIELDS]


class SqlMetricYmlParser(Parser):
    """
    Parses SQL metric yaml files
    """

    def __init__(self, sql_metric_dict: dict, sql_metric_path: str):
        super().__init__(description=sql_metric_path)

        if isinstance(sql_metric_dict, dict):
            self._push_context(object=sql_metric_dict, name=self.description)

            sql_metric_path_dir, sql_metric_file_name = FileSystemSingleton.INSTANCE.split(sql_metric_path)

            metric_names = self.get_list_optional(KEY_METRIC_NAMES)
            group_fields = self.get_list_optional(KEY_GROUP_FIELDS)
            sql = self.get_str_required(KEY_SQL)
            tests = self.parse_tests(
                sql_metric_dict,
                KEY_TESTS,
                context_sql_metric=sql_metric_file_name)

            self.sql_metric: SqlMetricYml = SqlMetricYml(sql=sql,
                                                         metric_names=metric_names,
                                                         file_name=sql_metric_file_name,
                                                         group_fields=group_fields,
                                                         tests=tests)

            self.check_invalid_keys(VALID_SQL_METRIC_KEYS)


        else:
            self.error('No SQL metric configuration provided')



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from sodasql.scan.dialect import Dialect
from sodasql.scan.parser import Parser


class DialectParser(Parser):

    def __init__(self, warehouse_connection_dict: dict):
        super().__init__('warehouse')
        self._push_context(warehouse_connection_dict, 'connection')
        self.dialect = Dialect.create(self)



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
from datetime import datetime
from typing import List


def sql_fetchone(connection, sql: str) -> tuple:
    """
    Only returns the tuple obtained by cursor.fetchone()
    """
    return sql_fetchone_description(connection, sql)[0]


def sql_fetchone_description(connection, sql: str) -> tuple:
    """
    Returns a tuple with 2 elements:
    1) the tuple obtained by cursor.fetchone()
    2) the cursor.description
    """
    cursor = connection.cursor()
    try:
        logging.debug(f'Executing SQL query: \n{sql}')
        start = datetime.now()
        cursor.execute(sql)
        row_tuple = cursor.fetchone()
        description = cursor.description
        delta = datetime.now() - start
        logging.debug(f'SQL took {str(delta)}')
        return row_tuple, description
    finally:
        cursor.close()


def sql_fetchall(connection, sql: str) -> List[tuple]:
    """
    Only returns the tuples obtained by cursor.fetchall()
    """
    return sql_fetchall_description(connection, sql)[0]


def sql_fetchall_description(connection, sql: str) -> tuple:
    """
    Returns a tuple with 2 elements:
    1) the tuples obtained by cursor.fetchall()
    2) the cursor.description
    """
    cursor = connection.cursor()
    try:
        logging.debug(f'Executing SQL query: \n{sql}')
        start = datetime.now()
        cursor.execute(sql)
        rows = cursor.fetchall()
        delta = datetime.now() - start
        logging.debug(f'SQL took {str(delta)}')
        return rows, cursor.description
    finally:
        cursor.close()


def sql_update(connection, sql: str):
    cursor = connection.cursor()
    try:
        logging.debug(f'Executing SQL update: \n{sql}')
        start = datetime.now()
        cursor.execute(sql)
        delta = datetime.now() - start
        logging.debug(f'SQL took {str(delta)}')
    finally:
        cursor.close()


def sql_updates(connection, sqls: List[str]):
    for sql in sqls:
        sql_update(connection, sql)



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


class Metric:

    ROW_COUNT = 'row_count'
    SCHEMA = 'schema'

    MISSING_COUNT = 'missing_count'
    MISSING_PERCENTAGE = 'missing_percentage'
    VALUES_COUNT = 'values_count'
    VALUES_PERCENTAGE = 'values_percentage'
    VALID_COUNT = 'valid_count'
    VALID_PERCENTAGE = 'valid_percentage'
    INVALID_COUNT = 'invalid_count'
    INVALID_PERCENTAGE = 'invalid_percentage'
    MIN = 'min'
    MAX = 'max'
    AVG = 'avg'
    SUM = 'sum'
    VARIANCE = 'variance'
    STDDEV = 'stddev'
    MIN_LENGTH = 'min_length'
    MAX_LENGTH = 'max_length'
    AVG_LENGTH = 'avg_length'
    DISTINCT = 'distinct'
    UNIQUE_COUNT = 'unique_count'
    DUPLICATE_COUNT = 'duplicate_count'
    UNIQUENESS = 'uniqueness'
    MAXS = 'maxs'
    MINS = 'mins'
    FREQUENT_VALUES = 'frequent_values'
    HISTOGRAM = 'histogram'

    METRIC_GROUP_MISSING = 'missing'
    METRIC_GROUP_VALIDITY = 'validity'
    METRIC_GROUP_DUPLICATES = 'duplicates'
    METRIC_GROUP_STATISTICS = 'statistics'
    METRIC_GROUP_LENGTH = 'length'
    METRIC_GROUP_PROFILING = 'profiling'

    METRIC_GROUPS = {
        METRIC_GROUP_MISSING: [MISSING_COUNT, MISSING_PERCENTAGE, VALUES_COUNT, VALUES_PERCENTAGE],
        METRIC_GROUP_VALIDITY: [VALID_COUNT, VALID_PERCENTAGE, INVALID_COUNT, INVALID_PERCENTAGE],
        METRIC_GROUP_DUPLICATES: [DISTINCT, UNIQUE_COUNT, UNIQUENESS, DUPLICATE_COUNT],
        METRIC_GROUP_STATISTICS: [MIN, MAX, AVG, SUM, VARIANCE, STDDEV],
        METRIC_GROUP_LENGTH: [MIN_LENGTH, MAX_LENGTH, AVG_LENGTH],
        METRIC_GROUP_PROFILING: [MAXS, MINS, FREQUENT_VALUES, HISTOGRAM]
    }

    METRIC_TYPES = [
        ROW_COUNT,
        SCHEMA,
        MISSING_COUNT,
        MISSING_PERCENTAGE,
        VALUES_COUNT,
        VALUES_PERCENTAGE,
        VALID_COUNT,
        VALID_PERCENTAGE,
        INVALID_COUNT,
        INVALID_PERCENTAGE,
        MIN,
        MAX,
        AVG,
        SUM,
        VARIANCE,
        STDDEV,
        MIN_LENGTH,
        MAX_LENGTH,
        AVG_LENGTH,
        DISTINCT,
        UNIQUE_COUNT,
        UNIQUENESS,
        DUPLICATE_COUNT,
        MAXS,
        MINS,
        FREQUENT_VALUES,
        HISTOGRAM
    ]



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import yaml


class IndentingDumper(yaml.Dumper):
    """
    yaml.dump hack to get indentation.
    see also https://stackoverflow.com/questions/25108581/python-yaml-dump-bad-indentation
    """

    def increase_indent(self, flow=False, indentless=False):
        return super(IndentingDumper, self).increase_indent(flow, False)





#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import List


class MoneyPatternHelper:

    @staticmethod
    def currency_number_pattern(thousands_separator, decimal_separator):
        return rf"(\d+[{thousands_separator}])*(\d+)({decimal_separator}\d+)?"

    @staticmethod
    def currency_prefix_pattern(currency_symbol):
        return rf"({currency_symbol})? ?(\-)?"

    @staticmethod
    def currency_suffix_pattern(currency_symbol, currency_name):
        return rf" ?({currency_symbol}|{currency_name.lower()}|{currency_name.upper()})?"

    @classmethod
    def money_pattern(cls, thousands_separator, decimal_separator, currency_symbol, currency_name):
        return cls.currency_prefix_pattern(currency_symbol) + \
               cls.currency_number_pattern(thousands_separator, decimal_separator, ) + \
               cls.currency_suffix_pattern(currency_symbol, currency_name)

    @staticmethod
    def or_patterns(patterns: List[str]):
        return '(' + '|'.join([f"({pattern})" for pattern in patterns]) + ')'

    @staticmethod
    def enclose_pattern(pattern):
        return r'^' + pattern + r'$'


MONEY_USD_PATTERN = MoneyPatternHelper.money_pattern(r',', r'\.', r'\$', r'usd')
MONEY_EUR_PATTERN = MoneyPatternHelper.money_pattern(r' ', r',', r'€', r'eur')
MONEY_GBP_PATTERN = MoneyPatternHelper.money_pattern(r',', r'\.', r'£', r'gbp')
MONEY_RMB_PATTERN = MoneyPatternHelper.money_pattern(r',', r'\.', r'¥', r'rmb')
MONEY_CHF_PATTERN = MoneyPatternHelper.money_pattern(r"''", r'\.', r'CHf', r'chf')



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import json
import logging
from typing import Optional

import requests

from sodasql.version import SODA_SQL_VERSION


class SodaServerClient:

    def __init__(self,
                 host: str,
                 port: Optional[str] = None,
                 protocol: Optional[str] = 'https',
                 username: Optional[str] = None,
                 password: Optional[str] = None,
                 api_key_id: Optional[str] = None,
                 api_key_secret: Optional[str] = None,
                 token: Optional[str] = None):
        self.host: str = host
        colon_port = f':{port}' if port else ''
        self.api_url: str = f'{protocol}://{self.host}{colon_port}/api'
        self.username: Optional[str] = username
        self.password: Optional[str] = password
        self.api_key_id: Optional[str] = api_key_id
        self.api_key_secret: Optional[str] = api_key_secret
        self.token: Optional[str] = token

    def scan_start(self, warehouse_name, warehouse_type, table_name, scan_time):
        return self.execute_command({
            'type': 'sodaSqlScanStart',
            'warehouseName': warehouse_name,
            'warehouseType': warehouse_type,
            'tableName': table_name,
            'scanTime': scan_time
            # 'columns': {
            #     'ID': {
            #         'missingValues': [ 'N/A', 'No value' ],
            #         'validity': {
            #               'namedFormat': 'date_eu',
            #               'regexFormat': 'someregex*',
            #               'allowedValues': ['a', 5],
            #               'minLength': 5,
            #               'maxLength': 25,
            #               'minValue': 0,
            #               'maxValue': 200
            #         }
            #     }
            # }
        })

    def scan_ended(self, scan_reference, exception = None):
        if exception is None:
            self.execute_command({
                'type': 'sodaSqlScanEnd',
                'scanReference': scan_reference
            })
        else:
            self.execute_command({
                'type': 'sodaSqlScanEnd',
                'scanReference': scan_reference,
                'error': str(exception)
            })

    def scan_measurements(self, scan_reference: dict, measurement_jsons: list):
        return self.execute_command({
            'type': 'sodaSqlScanMeasurements',
            'scanReference': scan_reference,
            'measurements': measurement_jsons
        })

    def scan_test_results(self, scan_reference: dict, test_result_jsons: list):
        return self.execute_command({
            'type': 'sodaSqlScanTestResults',
            'scanReference': scan_reference,
            'testResults': test_result_jsons
        })

    def execute_command(self, command: dict):
        return self._execute_request('command', command, False)

    def execute_query(self, command: dict):
        return self._execute_request('query', command, False)

    def _execute_request(self, request_type: str, request_body: dict, is_retry: bool):
        logging.debug(f'> /api/{request_type} {json.dumps(request_body, indent=2)}')
        request_body['token'] = self.get_token()
        request_body['sodaSqlVersion'] = SODA_SQL_VERSION
        response = requests.post(f'{self.api_url}/{request_type}', json=request_body)
        response_json = response.json()
        logging.debug(f'< {response.status_code} {json.dumps(response_json, indent=2)}')
        if response.status_code == 401 and not is_retry:
            logging.debug(f'Authentication failed. Probably token expired. Reauthenticating...')
            self.token = None
            response_json = self._execute_request(request_type, request_body, True)
        else:
            assert response.status_code == 200, f'Request failed with status {response.status_code}: {json.dumps(response_json, indent=2)}'
        return response_json

    def get_token(self):
        if not self.token:
            login_command = {
                'type': 'login'
            }
            if self.api_key_id and self.api_key_secret:
                logging.debug('> /api/command (login with API key credentials)')
                login_command['apiKeyId'] = self.api_key_id
                login_command['apiKeySecret'] = self.api_key_secret
            elif self.username and self.password:
                logging.debug('> /api/command (login with username and password)')
                login_command['username'] = self.username
                login_command['password'] = self.password
            else:
                raise RuntimeError('No authentication in environment variables')

            login_response = requests.post(f'{self.api_url}/command', json=login_command)

            if login_response.status_code != 200:
                raise AssertionError(f'< {login_response.status_code} Login failed: {login_response.content}')
            login_response_json = login_response.json()
            self.token = login_response_json.get('token')
            assert self.token, 'No token in login response?!'
            logging.debug('< 200 (login ok, token received)')
        return self.token



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from dataclasses import dataclass
from typing import Optional, List

from sodasql.common.json_helper import JsonHelper
from sodasql.scan.group_value import GroupValue


@dataclass
class Measurement:

    metric: str
    column_name: Optional[str] = None
    value: object = None
    group_values: Optional[List[GroupValue]] = None

    def __str__(self):
        column_str = f'({self.column_name})' if self.column_name else ''
        group_values_str = JsonHelper.to_json(JsonHelper.to_jsonnable(self.group_values)) if self.group_values else ''
        value_str = '' if self.value is None \
                    else (' = ' + (', '.join([str(e) for e in self.value]) if isinstance(self.value, list)
                                  else str(self.value)))
        return f'{self.metric}{column_str}{group_values_str}{value_str}'

    def to_json(self):
        json = {
            'metric': self.metric,
        }

        if self.group_values is None:
            json['value'] = JsonHelper.to_jsonnable(self.value)
        else:
            json['groupValues'] = [group_value.to_json() for group_value in self.group_values]

        if self.column_name is not None:
            json['columnName'] = self.column_name

        return json



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
from dataclasses import dataclass
from typing import Optional


@dataclass
class Test:

    description: str
    expression: str
    first_metric: str
    column: Optional[str]

    def evaluate(self, test_variables: dict, group_values: Optional[dict] = None):
        from sodasql.scan.test_result import TestResult
        try:
            passed = bool(eval(self.expression, test_variables))
            value = test_variables.get(self.first_metric)
            logging.debug(f'Test {self.description} {"passed" if passed else "failed"}' +
                          (f" with group values {group_values}" if group_values else ''))

            return TestResult(test=self, passed=passed, value=value, group_values=group_values)
        except Exception as e:
            return TestResult(test=self, passed=False, error=str(e), group_values=group_values)



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
import re
from math import ceil

import yaml

from sodasql.scan.file_system import FileSystemSingleton
from sodasql.cli.indenting_yaml_dumper import IndentingDumper
from sodasql.scan.metric import Metric
from sodasql.scan.validity import Validity
from sodasql.scan.warehouse import Warehouse


class ScanInitializer:

    def __init__(self, warehouse: Warehouse, warehouse_dir: str):
        self.warehouse: Warehouse = warehouse
        self.warehouse_dir: str = warehouse_dir
        self.first_table_scan_yml_file = None

    def initialize_scan_ymls(self):
        file_system = FileSystemSingleton.INSTANCE

        def fileify(name: str):
            return re.sub(r'[^A-Za-z0-9_.]+', '_', name).lower()

        table_dir = file_system.join(self.warehouse_dir, 'tables')
        if not file_system.file_exists(table_dir):
            logging.info(f'Creating tables directory {table_dir}')
            file_system.mkdirs(table_dir)
        else:
            logging.info(f'Directory {table_dir} already exists')

        rows = self.warehouse.sql_fetchall(
            self.warehouse.dialect.sql_tables_metadata_query())
        for row in rows:
            table_name = row[0]

            table_scan_yaml_file = file_system.join(table_dir, f'{fileify(table_name)}.yml')

            if not self.first_table_scan_yml_file:
                self.first_table_scan_yml_file = table_scan_yaml_file

            if file_system.file_exists(table_scan_yaml_file):
                logging.info(
                    f"Scan file {table_scan_yaml_file} already exists")
            else:
                logging.info(f"Creating {table_scan_yaml_file} ...")
                from sodasql.scan.scan_yml_parser import (KEY_METRICS,
                                                          KEY_TABLE_NAME,
                                                          KEY_TESTS,
                                                          KEY_COLUMNS,
                                                          COLUMN_KEY_VALID_FORMAT,
                                                          COLUMN_KEY_TESTS)
                scan_yaml_dict = {
                    KEY_TABLE_NAME: table_name,
                    KEY_METRICS:
                        [Metric.ROW_COUNT] +
                        Metric.METRIC_GROUPS[Metric.METRIC_GROUP_MISSING] +
                        Metric.METRIC_GROUPS[Metric.METRIC_GROUP_VALIDITY] +
                        Metric.METRIC_GROUPS[Metric.METRIC_GROUP_LENGTH] +
                        Metric.METRIC_GROUPS[Metric.METRIC_GROUP_STATISTICS],
                    KEY_TESTS: [
                        'row_count > 0'
                    ]
                }

                dialect = self.warehouse.dialect
                sql = dialect.sql_columns_metadata_query(table_name)
                column_tuples = self.warehouse.sql_fetchall(sql)
                qualified_table_name = dialect.qualify_table_name(table_name)
                columns = {}
                for column_tuple in column_tuples:
                    column_name = column_tuple[0]
                    column_type = column_tuple[1]
                    qualified_column_name = dialect.qualify_column_name(column_name)
                    if dialect.is_text(column_type):
                        validity_format_count_fields = []
                        validity_counts = []
                        for validity_format in Validity.FORMATS:
                            format_regex = Validity.FORMATS[validity_format]
                            validity_counts.append({'format': validity_format})
                            qualified_regex = dialect.qualify_regex(format_regex)
                            regexp_like = dialect.sql_expr_regexp_like(qualified_column_name, qualified_regex)
                            count_field = f'COUNT(CASE WHEN {regexp_like} THEN 1 END)'
                            validity_format_count_fields.append(count_field)

                        row = self.warehouse.sql_fetchone(
                            f'SELECT \n  ' +
                            (',\n  '.join(validity_format_count_fields)) + ',\n'
                            f'  COUNT({qualified_column_name}) \n'
                            f'FROM {qualified_table_name} \n'
                            f'LIMIT 1000'
                        )

                        values_count = row[len(validity_counts)]

                        if values_count > 0:
                            for i in range(len(validity_counts)):
                                validity_count = validity_counts[i]
                                validity_count['count'] = row[i]

                            sorted_validity_counts = sorted(validity_counts, key=lambda c: c['count'], reverse=True)
                            most_frequent_validity_format = sorted_validity_counts[0]
                            valid_count = most_frequent_validity_format['count']

                            if valid_count > (values_count / 2):
                                column_yml = {
                                    COLUMN_KEY_VALID_FORMAT: most_frequent_validity_format['format']
                                }
                                if valid_count > (values_count * .8):
                                    valid_percentage = valid_count * 100 / values_count
                                    invalid_threshold = (100 - valid_percentage) * 1.1
                                    invalid_threshold_rounded = ceil(invalid_threshold)
                                    column_yml[COLUMN_KEY_TESTS] = [
                                        f'invalid_percentage == {str(invalid_threshold_rounded)}'
                                    ]
                                columns[column_name] = column_yml

                if columns:
                    scan_yaml_dict[KEY_COLUMNS] = columns

                scan_yml_str = yaml.dump(scan_yaml_dict,
                                         sort_keys=False,
                                         Dumper=IndentingDumper,
                                         default_flow_style=False)
                file_system.file_write_from_str(table_scan_yaml_file, scan_yml_str)



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from dataclasses import dataclass
from typing import Optional

from sodasql.scan.dialect import Dialect
from sodasql.soda_server_client.soda_server_client import SodaServerClient


@dataclass
class WarehouseYml:

    dialect: Dialect = None
    name: str = None
    soda_host: Optional[str] = None
    soda_port: Optional[int] = None
    soda_protocol: Optional[str] = None
    soda_api_key_id: Optional[str] = None
    soda_api_key_secret: Optional[str] = None



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

class ColumnMetadata:

    def __init__(self, name: str, type: str = None, nullable: bool = None):
        self.name = name
        self.type = type
        self.nullable = nullable

    def __str__(self):
        return self.name+(' '+self.type if self.type else '')

    def to_json(self):
        return {
            'name': self.name,
            'type': self.type
        }



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from sodasql.scan.validity.money_patterns import MoneyPatternHelper
from sodasql.scan.validity.money_patterns import MONEY_USD_PATTERN, MONEY_EUR_PATTERN, MONEY_GBP_PATTERN, \
    MONEY_RMB_PATTERN, MONEY_CHF_PATTERN

class Validity:
    """
    For the sake of compatibility with all warehouses, we only support POSIX regexes. POSIX does not support ERE.
    This source makes the distinction between RE and ERE (Extended Regular Expression):
        http://pubs.opengroup.org/onlinepubs/9699919799/basedefs/V1_chap09.html#tag_09_05
    More regexes:
        https://digitalfortress.tech/tricks/top-15-commonly-used-regex/
        http://regexlib.com/
    """

    FORMATS = {
        'number_whole': r'^\-?\d+$',
        'number_decimal_point': r'^\-?\d+\.\d+$',
        'number_decimal_comma': r'^\-?\d+,\d+$',
        'number_percentage': r'^\-?\d+([\.,]\d+)? ?%$',
        'number_money_usd': MoneyPatternHelper.enclose_pattern(MONEY_USD_PATTERN),
        'number_money_eur': MoneyPatternHelper.enclose_pattern(MONEY_EUR_PATTERN),
        'number_money_gbp': MoneyPatternHelper.enclose_pattern(MONEY_GBP_PATTERN),
        'number_money_rmb': MoneyPatternHelper.enclose_pattern(MONEY_RMB_PATTERN),
        'number_money_chf': MoneyPatternHelper.enclose_pattern(MONEY_CHF_PATTERN),
        'number_money': MoneyPatternHelper.enclose_pattern(
            MoneyPatternHelper.or_patterns([MONEY_USD_PATTERN,
                                            MONEY_EUR_PATTERN,
                                            MONEY_GBP_PATTERN,
                                            MONEY_RMB_PATTERN,
                                            MONEY_CHF_PATTERN])
        ),

        'date_eu': r'^([1-9]|0[1-9]|[12][0-9]|3[01])[-\./]([1-9]|0[1-9]|1[012])[-\./](19|20)?\d\d',
        'date_us': r'^([1-9]|0[1-9]|1[012])[-\./]([1-9]|0[1-9]|[12][0-9]|3[01])[-\./](19|20)?\d\d',
        'date_inverse': r'^(19|20)\d\d[-\./]?([1-9]|0[1-9]|1[012])[-\./]?([1-9]|0[1-9]|[12][0-9]|3[01])',
        'time': r'([0-9]|1[0-9]|2[0-4])[:-]([0-9]|[0-5][0-9])([:-]([0-9]|[0-5][0-9])(,\d+)?)?$',
        'date_iso_8601':
            r'^'
            r'([1-9]\d{3}-((0[1-9]|1[0-2])-(0[1-9]|1\d|2[0-8])|(0[13-9]|1[0-2])-(29|30)|(0[13578]|1[02])-31)|'
            r'([1-9]\d(0[48]|[2468][048]|[13579][26])|([2468][048]|[13579][26])00)-02-29)'

            r'T([01]\d|2[0-3]):[0-5]\d:[0-5]\d(\.\d+)?'

            r'(Z|[+-][01]\d:[0-5]\d)?'
            r'$',

        'uuid': r'^[0-9a-fA-F]{8}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{4}\-[0-9a-fA-F]{12}$',
        'ip_address': r'^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$',

        'email': r'^[A-Za-z0-9.-_%]+@[A-Za-z0-9.-]+\.[A-Za-z]{2,4}$',
        'phone_number': r'^((\+\d{1,2}\s)?\(?\d{3}\)?[\s.-])?\d{3}[\s.-]\d{4}$',
        'credit_card_number': r'^\d{16}|(\d{4}-){3}\d{4}|(\d{4} ){3}\d{4}$',
    }

    def __init__(self):
        self.format = None
        self.regex = None
        self.values = None
        self.min_length = None
        self.max_length = None
        self.min = None
        self.max = None



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging

from sodasql.common.yaml_helper import YamlHelper
from sodasql.scan.dialect import Dialect
from sodasql.scan.env_vars import EnvVars
from sodasql.scan.file_system import FileSystemSingleton
from sodasql.scan.parser import Parser
from sodasql.scan.warehouse_yml import WarehouseYml

KEY_NAME = 'name'
KEY_CONNECTION = 'connection'
KEY_SODA_ACCOUNT = 'soda_account'

SODA_KEY_HOST = 'host'
SODA_KEY_PORT = 'port'
SODA_KEY_PROTOCOL = 'protocol'
SODA_KEY_API_KEY_ID = 'api_key_id'
SODA_KEY_API_KEY_SECRET = 'api_key_secret'

VALID_WAREHOUSE_KEYS = [KEY_NAME, KEY_CONNECTION, KEY_SODA_ACCOUNT]


def read_warehouse_yml_file(warehouse_yml_file: str):
    file_system = FileSystemSingleton.INSTANCE
    if file_system.is_readable_file(warehouse_yml_file):
        warehouse_yaml_str = file_system.file_read_as_str(warehouse_yml_file)
        if warehouse_yaml_str:
            return YamlHelper.parse_yaml(warehouse_yaml_str, warehouse_yml_file)
        else:
            logging.error(f'Failed to read warehouse yaml file: {warehouse_yml_file}')


class WarehouseYmlParser(Parser):
    """
    Parses warehouse.yml files
    """

    def __init__(self,
                 warehouse_dict: dict,
                 warehouse_yaml_file: str = 'warehouse-dict'):
        super().__init__(description=warehouse_yaml_file)

        if isinstance(warehouse_dict, dict):
            self._push_context(object=warehouse_dict, name=self.description)

            self.warehouse_yml = WarehouseYml()
            self.warehouse_yml.name = self.get_str_required(KEY_NAME)

            EnvVars.load_env_vars(self.warehouse_yml.name)

            connection_dict = self.get_dict_required(KEY_CONNECTION)
            if connection_dict:
                self._push_context(object=connection_dict, name=KEY_CONNECTION)
                self.warehouse_yml.dialect = Dialect.create(self)
                self._pop_context()

            soda_account_dict = self.get_dict_optional(KEY_SODA_ACCOUNT)
            if soda_account_dict:
                self._push_context(object=soda_account_dict, name=KEY_SODA_ACCOUNT)
                self.warehouse_yml.soda_host = self.get_str_optional(SODA_KEY_HOST, 'cloud.soda.io')
                self.warehouse_yml.soda_port = self.get_int_optional(SODA_KEY_PORT, 443)
                self.warehouse_yml.soda_protocol = self.get_str_optional(SODA_KEY_PROTOCOL, 'https')
                self.warehouse_yml.soda_api_key_id = self.get_str_required_env(SODA_KEY_API_KEY_ID)
                self.warehouse_yml.soda_api_key_secret = self.get_str_required_env(SODA_KEY_API_KEY_SECRET)
                self._pop_context()

            self.check_invalid_keys(VALID_WAREHOUSE_KEYS)

        else:
            self.error('No warehouse configuration provided')



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
import traceback
from math import floor, ceil
from typing import List, Optional

from jinja2 import Template

from sodasql.scan.column_metadata import ColumnMetadata
from sodasql.scan.group_value import GroupValue
from sodasql.scan.measurement import Measurement
from sodasql.scan.metric import Metric
from sodasql.scan.scan_column import ScanColumn
from sodasql.scan.scan_result import ScanResult
from sodasql.scan.scan_yml import ScanYml
from sodasql.scan.sql_metric_yml import SqlMetricYml
from sodasql.scan.test_result import TestResult
from sodasql.scan.warehouse import Warehouse
from sodasql.soda_server_client.soda_server_client import SodaServerClient


class Scan:

    def __init__(self,
                 warehouse: Warehouse,
                 scan_yml: ScanYml = None,
                 soda_server_client: SodaServerClient = None,
                 variables: dict = None,
                 time: str = None):
        self.warehouse = warehouse
        self.scan_yml = scan_yml
        self.soda_server_client = soda_server_client
        self.variables = variables
        self.time = time

        self.scan_result = ScanResult()
        self.dialect = warehouse.dialect
        self.qualified_table_name = self.dialect.qualify_table_name(scan_yml.table_name)
        self.scan_reference = None
        self.column_metadatas: List[ColumnMetadata] = []
        self.column_names: List[str] = []
        # maps column names (lower case) to ScanColumn's
        self.scan_columns: dict = {}
        self.close_warehouse = True
        self.send_scan_end = True
        self.exception = None

        self.table_sample_clause = \
            f'\nTABLESAMPLE {scan_yml.sample_method}({scan_yml.sample_percentage})' \
            if scan_yml.sample_percentage \
            else ''

        self.filter_sql = None
        if scan_yml.filter_template:
            if not variables:
                raise RuntimeError(f'No variables provided while filter "{str(scan_yml.filter)}" specified')
            self.filter_sql = scan_yml.filter_template.render(variables)

    def execute(self) -> ScanResult:
        self._ensure_scan_reference()

        try:
            if self.scan_yml:
                # Soda Server and the code below require that the schema measurements is the first measurement
                self._query_columns_metadata()
                self._query_aggregations()
                self._query_group_by_value()
                self._query_histograms()

            self._query_sql_metrics_and_run_tests()
            self._run_table_tests()
            self._run_column_tests()

        except Exception as e:
            logging.exception('Scan failed')
            self.exception = e
            self.scan_result.error = traceback.format_exc()

        finally:
            try:
                if self.soda_server_client and self.send_scan_end:
                    self.soda_server_client.scan_ended(self.scan_reference, self.exception)
            except:
                logging.exception('Notifying Soda Server of scan ended failed')

            if self.close_warehouse:
                self.warehouse.close()

        return self.scan_result

    def _query_columns_metadata(self):
        sql = self.warehouse.dialect.sql_columns_metadata_query(self.scan_yml.table_name)
        column_tuples = self.warehouse.sql_fetchall(sql)
        self.column_metadatas = []
        for column_tuple in column_tuples:
            name = column_tuple[0]
            type = column_tuple[1]
            nullable = 'YES' == column_tuple[2].upper()
            self.column_metadatas.append(ColumnMetadata(name, type, nullable))
        logging.debug(str(len(self.column_metadatas)) + ' columns:')
        for column in self.column_metadatas:
            logging.debug(f'  {column.name} {column.type} {"" if column.nullable else "not null"}')

        self.column_names: List[str] = [column_metadata.name for column_metadata in self.column_metadatas]
        self.scan_columns: dict = {column.name.lower(): ScanColumn(self, column) for column in self.column_metadatas}
        schema_measurement_value = [column_metadata.to_json() for column_metadata in self.column_metadatas]
        schema_measurement = Measurement(Metric.SCHEMA, value=schema_measurement_value)
        self._log_measurement(schema_measurement)
        self._flush_measurements([schema_measurement])

    def _query_aggregations(self):
        # This measurements list is used to match measurements with the query field order.
        # After query execution, the value of the measurements will be extracted from the query result and
        # the measurements will be added with self.add_query(measurement)
        measurements: List[Measurement] = []

        fields: List[str] = []

        dialect = self.warehouse.dialect

        if self.scan_yml.is_metric_enabled(Metric.ROW_COUNT):
            fields.append(dialect.sql_expr_count_all())
            measurements.append(Measurement(Metric.ROW_COUNT))

        # maps db column names (lower) to missing and invalid metric indices in the measurements
        # eg { 'colname': {'missing': 2, 'invalid': 3}, ...}
        column_metric_indices = {}

        for column_name_lower in self.scan_columns:
            metric_indices = {}
            column_metric_indices[column_name_lower] = metric_indices
            scan_column: ScanColumn = self.scan_columns[column_name_lower]
            column_name = scan_column.column_name

            if scan_column.is_missing_enabled:
                metric_indices['non_missing'] = len(measurements)
                if scan_column.non_missing_condition:
                    fields.append(dialect.sql_expr_count_conditional(scan_column.non_missing_condition))
                else:
                    fields.append(dialect.sql_expr_count(scan_column.qualified_column_name))
                measurements.append(Measurement(Metric.VALUES_COUNT, column_name))

            if scan_column.is_valid_enabled:
                metric_indices['valid'] = len(measurements)
                if scan_column.non_missing_and_valid_condition:
                    fields.append(dialect.sql_expr_count_conditional(scan_column.non_missing_and_valid_condition))
                else:
                    fields.append(dialect.sql_expr_count(scan_column.qualified_column_name))
                measurements.append(Measurement(Metric.VALID_COUNT, column_name))

            if scan_column.is_text:
                length_expr = dialect.sql_expr_conditional(
                        scan_column.non_missing_and_valid_condition,
                        dialect.sql_expr_length(scan_column.qualified_column_name)) \
                    if scan_column.non_missing_and_valid_condition \
                    else dialect.sql_expr_length(scan_column.qualified_column_name)

                if self.scan_yml.is_metric_enabled(Metric.MIN_LENGTH, column_name):
                    fields.append(dialect.sql_expr_min(length_expr))
                    measurements.append(Measurement(Metric.MIN_LENGTH, column_name))

                if self.scan_yml.is_metric_enabled(Metric.MAX_LENGTH, column_name):
                    fields.append(dialect.sql_expr_max(length_expr))
                    measurements.append(Measurement(Metric.MAX_LENGTH, column_name))

            if scan_column.has_numeric_values:
                if scan_column.is_metric_enabled(Metric.MIN):
                    fields.append(dialect.sql_expr_min(scan_column.numeric_expr))
                    measurements.append(Measurement(Metric.MIN, column_name))

                if scan_column.is_metric_enabled(Metric.MAX):
                    fields.append(dialect.sql_expr_max(scan_column.numeric_expr))
                    measurements.append(Measurement(Metric.MAX, column_name))

                if scan_column.is_metric_enabled(Metric.AVG):
                    fields.append(dialect.sql_expr_avg(scan_column.numeric_expr))
                    measurements.append(Measurement(Metric.AVG, column_name))

                if scan_column.is_metric_enabled(Metric.SUM):
                    fields.append(dialect.sql_expr_sum(scan_column.numeric_expr))
                    measurements.append(Measurement(Metric.SUM, column_name))

                if scan_column.is_metric_enabled(Metric.VARIANCE):
                    fields.append(dialect.sql_expr_variance(scan_column.numeric_expr))
                    measurements.append(Measurement(Metric.VARIANCE, column_name))

                if scan_column.is_metric_enabled(Metric.STDDEV):
                    fields.append(dialect.sql_expr_stddev(scan_column.numeric_expr))
                    measurements.append(Measurement(Metric.STDDEV, column_name))

        if len(fields) > 0:
            sql = 'SELECT \n  ' + ',\n  '.join(fields) + ' \n' \
                  'FROM ' + self.qualified_table_name
            if self.table_sample_clause:
                sql += f'\n{self.table_sample_clause}'
            if self.filter_sql:
                sql += f'\nWHERE {self.filter_sql}'

            query_result_tuple = self.warehouse.sql_fetchone(sql)

            for i in range(0, len(measurements)):
                measurement = measurements[i]
                measurement.value = query_result_tuple[i]
                self._log_measurement(measurement)

            # Calculating derived measurements
            row_count_measurement = next((m for m in measurements if m.metric == Metric.ROW_COUNT), None)
            if row_count_measurement:
                row_count = row_count_measurement.value
                for column_name_lower in self.scan_columns:
                    scan_column = self.scan_columns[column_name_lower]
                    column_name = scan_column.column_name
                    metric_indices = column_metric_indices[column_name_lower]
                    non_missing_index = metric_indices.get('non_missing')
                    if non_missing_index is not None:
                        values_count = measurements[non_missing_index].value
                        missing_count = row_count - values_count
                        missing_percentage = missing_count * 100 / row_count if row_count > 0 else None
                        values_percentage = values_count * 100 / row_count if row_count > 0 else None

                        self._log_and_append_derived_measurements(measurements, [
                            Measurement(Metric.MISSING_PERCENTAGE, column_name, missing_percentage),
                            Measurement(Metric.MISSING_COUNT, column_name, missing_count),
                            Measurement(Metric.VALUES_PERCENTAGE, column_name, values_percentage)
                        ])

                        valid_index = metric_indices.get('valid')
                        if valid_index is not None:
                            valid_count = measurements[valid_index].value
                            invalid_count = row_count - missing_count - valid_count
                            invalid_percentage = invalid_count * 100 / row_count if row_count > 0 else None
                            valid_percentage = valid_count * 100 / row_count if row_count > 0 else None
                            self._log_and_append_derived_measurements(measurements, [
                                Measurement(Metric.INVALID_PERCENTAGE, column_name, invalid_percentage),
                                Measurement(Metric.INVALID_COUNT, column_name, invalid_count),
                                Measurement(Metric.VALID_PERCENTAGE, column_name, valid_percentage)
                            ])

        self._flush_measurements(measurements)

    def _query_group_by_value(self):
        for column_name_lower in self.scan_columns:
            measurements = []

            scan_column: ScanColumn = self.scan_columns[column_name_lower]
            column_name = scan_column.column_name

            if scan_column.is_any_metric_enabled(
                    [Metric.DISTINCT, Metric.UNIQUENESS, Metric.UNIQUE_COUNT,
                     Metric.MINS, Metric.MAXS, Metric.FREQUENT_VALUES, Metric.DUPLICATE_COUNT]):

                group_by_cte = scan_column.get_group_by_cte()
                numeric_value_expr = scan_column.get_group_by_cte_numeric_value_expression()

                if self.scan_yml.is_any_metric_enabled(
                        [Metric.DISTINCT, Metric.UNIQUENESS, Metric.UNIQUE_COUNT, Metric.DUPLICATE_COUNT],
                        column_name):

                    sql = (f'{group_by_cte} \n'
                           f'SELECT COUNT(*), \n'
                           f'       COUNT(CASE WHEN frequency = 1 THEN 1 END), \n'
                           f'       SUM(frequency) \n'
                           f'FROM group_by_value')

                    query_result_tuple = self.warehouse.sql_fetchone(sql)
                    distinct_count = query_result_tuple[0]
                    unique_count = query_result_tuple[1]
                    valid_count = query_result_tuple[2] if query_result_tuple[2] else 0
                    duplicate_count = distinct_count - unique_count

                    self._log_and_append_query_measurement(
                        measurements, Measurement(Metric.DISTINCT, column_name, distinct_count))
                    self._log_and_append_query_measurement(
                        measurements, Measurement(Metric.UNIQUE_COUNT, column_name, unique_count))

                    derived_measurements = [Measurement(Metric.DUPLICATE_COUNT, column_name, duplicate_count)]
                    if valid_count > 1:
                        uniqueness = (distinct_count - 1) * 100 / (valid_count - 1)
                        derived_measurements.append(Measurement(Metric.UNIQUENESS, column_name, uniqueness))
                    self._log_and_append_derived_measurements(measurements, derived_measurements)

                if scan_column.is_metric_enabled(Metric.MINS) and scan_column.numeric_expr is not None:

                    sql = (f'{group_by_cte} \n'
                           f'SELECT value \n'
                           f'FROM group_by_value \n'
                           f'ORDER BY {numeric_value_expr} ASC \n'
                           f'LIMIT {scan_column.mins_maxs_limit} \n')

                    rows = self.warehouse.sql_fetchall(sql)
                    mins = [row[0] for row in rows]
                    self._log_and_append_query_measurement(measurements, Measurement(Metric.MINS, column_name, mins))

                if self.scan_yml.is_metric_enabled(Metric.MAXS, column_name) \
                        and (scan_column.is_number or scan_column.is_column_numeric_text_format):

                    sql = (f'{group_by_cte} \n'
                           f'SELECT value \n'
                           f'FROM group_by_value \n'
                           f'ORDER BY {numeric_value_expr} DESC \n'
                           f'LIMIT {scan_column.mins_maxs_limit} \n')

                    rows = self.warehouse.sql_fetchall(sql)
                    maxs = [row[0] for row in rows]
                    self._log_and_append_query_measurement(measurements, Measurement(Metric.MAXS, column_name, maxs))

                if self.scan_yml.is_metric_enabled(Metric.FREQUENT_VALUES, column_name) \
                        and (scan_column.is_number or scan_column.is_column_numeric_text_format):

                    frequent_values_limit = self.scan_yml.get_frequent_values_limit(column_name)
                    sql = (f'{group_by_cte} \n'
                           f'SELECT value, frequency \n'
                           f'FROM group_by_value \n'
                           f'ORDER BY frequency DESC \n'
                           f'LIMIT {frequent_values_limit} \n')

                    rows = self.warehouse.sql_fetchall(sql)
                    frequent_values = [{'value': row[0], 'frequency': row[1]} for row in rows]
                    self._log_and_append_query_measurement(
                        measurements, Measurement(Metric.FREQUENT_VALUES, column_name, frequent_values))

            self._flush_measurements(measurements)

    def _query_histograms(self):
        measurements = []
        for column_name_lower in self.scan_columns:
            scan_column: ScanColumn = self.scan_columns[column_name_lower]
            column_name = scan_column.column_name

            if scan_column.is_metric_enabled(Metric.HISTOGRAM) and scan_column.numeric_expr:

                buckets: int = scan_column.get_histogram_buckets()

                min_value = scan_column.get_metric_value(Metric.MIN)
                max_value = scan_column.get_metric_value(Metric.MAX)

                if scan_column.has_numeric_values and min_value and max_value and min_value < max_value:
                    # Build the histogram query
                    min_value = floor(min_value * 1000) / 1000
                    max_value = ceil(max_value * 1000) / 1000
                    bucket_width = (max_value - min_value) / buckets

                    boundary = min_value
                    boundaries = [min_value]
                    for i in range(0, buckets):
                        boundary += bucket_width
                        boundaries.append(round(boundary, 3))

                    group_by_cte = scan_column.get_group_by_cte()
                    numeric_value_expr = scan_column.get_group_by_cte_numeric_value_expression()

                    field_clauses = []
                    for i in range(0, buckets):
                        lower_bound = '' if i == 0 else f'{boundaries[i]} <= {numeric_value_expr}'
                        upper_bound = '' if i == buckets - 1 else f'{numeric_value_expr} < {boundaries[i + 1]}'
                        optional_and = '' if lower_bound == '' or upper_bound == '' else ' and '
                        field_clauses.append(
                            f'SUM(CASE WHEN {lower_bound}{optional_and}{upper_bound} THEN frequency END)')

                    fields = ',\n  '.join(field_clauses)

                    sql = (f'{group_by_cte} \n'
                           f'SELECT \n'
                           f'  {fields} \n'
                           f'FROM group_by_value')

                    if self.filter_sql:
                        sql += f' \nWHERE {self.scan.filter_sql}'

                    row = self.warehouse.sql_fetchone(sql)

                    # Process the histogram query
                    frequencies = []
                    for i in range(0, buckets):
                        frequency = row[i]
                        frequencies.append(0 if not frequency else int(frequency))
                    histogram = {
                        'boundaries': boundaries,
                        'frequencies': frequencies
                    }

                    self._log_and_append_query_measurement(
                        measurements, Measurement(Metric.HISTOGRAM, column_name, histogram))
        self._flush_measurements(measurements)

    def _query_sql_metrics_and_run_tests(self):
        self._query_sql_metrics_and_run_tests_base(self.scan_yml.sql_metric_ymls)
        for column_name_lower in self.scan_columns:
            scan_column: ScanColumn = self.scan_columns[column_name_lower]
            if scan_column and scan_column.scan_yml_column:
                self._query_sql_metrics_and_run_tests_base(scan_column.scan_yml_column.sql_metric_ymls, column_name_lower)

    def _query_sql_metrics_and_run_tests_base(self, sql_metric_ymls: Optional[List[SqlMetricYml]], column_name_lower: Optional[str] = None):
        if sql_metric_ymls:
            for sql_metric in sql_metric_ymls:
                if self.variables:
                    sql_variables = self.variables.copy() if self.variables else {}
                    # TODO add functions to sql_variables that can convert scan variables to valid SQL literals
                    template = Template(sql_metric.sql)
                    resolved_sql = template.render(sql_variables)
                else:
                    resolved_sql = sql_metric.sql

                if sql_metric.group_fields:
                    self._run_sql_metric_with_groups_and_run_tests(sql_metric, resolved_sql, column_name_lower)
                else:
                    self._run_sql_metric_default_and_run_tests(sql_metric, resolved_sql, column_name_lower)

    def _run_sql_metric_default_and_run_tests(self, sql_metric: SqlMetricYml, resolved_sql: str, column_name_lower: Optional[str] = None):
        row_tuple, description = self.warehouse.sql_fetchone_description(resolved_sql)
        test_variables = self._get_test_variables(column_name_lower)

        measurements = []
        for i in range(len(row_tuple)):
            metric_name = sql_metric.metric_names[i] if sql_metric.metric_names is not None else description[i][0]
            metric_value = row_tuple[i]
            logging.debug(f'SQL metric {sql_metric.name} {metric_name} -> {metric_value}')
            measurement = Measurement(metric=metric_name, value=metric_value)
            test_variables[metric_name] = metric_value
            self._log_and_append_query_measurement(measurements, measurement)

        self._flush_measurements(measurements)

        sql_metric_test_results = self._execute_tests(sql_metric.tests, test_variables)
        self._flush_test_results(sql_metric_test_results)

    def _run_sql_metric_with_groups_and_run_tests(self, sql_metric: SqlMetricYml, resolved_sql: str, column_name_lower: Optional[str] = None):
        measurements = []
        test_results = []
        group_fields_lower = set(group_field.lower() for group_field in sql_metric.group_fields)
        rows, description = self.warehouse.sql_fetchall_description(resolved_sql)
        group_values_by_metric_name = {}
        for row in rows:
            group = {}
            metric_values = {}
            for i in range(len(row)):
                metric_name = sql_metric.metric_names[i] if sql_metric.metric_names is not None else description[i][0]
                metric_value = row[i]
                if metric_name.lower() in group_fields_lower:
                    group[metric_name] = metric_value
                else:
                    metric_values[metric_name] = metric_value

            for metric_name in metric_values:
                metric_value = metric_values[metric_name]
                if metric_name not in group_values_by_metric_name:
                    group_values_by_metric_name[metric_name] = []
                group_values = group_values_by_metric_name[metric_name]
                group_values.append(GroupValue(group=group, value=metric_value))
                logging.debug(f'SQL metric {sql_metric.name} {metric_name} {group} -> {metric_value}')

            sql_metric_tests = sql_metric.tests
            test_variables = self._get_test_variables(column_name_lower)
            test_variables.update(metric_values)
            sql_metric_test_results = self._execute_tests(sql_metric_tests, test_variables, group)
            test_results.extend(sql_metric_test_results)

        for metric_name in group_values_by_metric_name:
            group_values = group_values_by_metric_name[metric_name]
            measurement = Measurement(metric=metric_name, group_values=group_values)
            self._log_and_append_query_measurement(measurements, measurement)

        self._flush_measurements(measurements)
        self._flush_test_results(test_results)

    def _get_test_variables(self, column_name_lower: Optional[str] = None):
        return {
            measurement.metric: measurement.value
            for measurement in self.scan_result.measurements
            if measurement.column_name is None or measurement.column_name.lower() == column_name_lower
        }

    def _run_table_tests(self):
        test_variables = self._get_test_variables()

        table_tests = self.scan_yml.tests
        table_test_results = self._execute_tests(table_tests, test_variables)
        self._flush_test_results(table_test_results)

    def _run_column_tests(self):
        test_results = []
        for column_name_lower in self.scan_columns:
            scan_column: ScanColumn = self.scan_columns[column_name_lower]
            column_tests = scan_column.get_tests()
            test_variables = self._get_test_variables(column_name_lower)
            column_test_results = self._execute_tests(column_tests, test_variables)
            test_results.extend(column_test_results)
        self._flush_test_results(test_results)

    def _execute_tests(self, tests, variables, group_values: Optional[dict] = None):
        test_results = []
        if tests:
            for test in tests:
                test_results.append(test.evaluate(variables, group_values))
        return test_results

    @classmethod
    def _log_and_append_query_measurement(cls, measurements: List[Measurement], measurement: Measurement):
        """
        Convenience method to log a measurement and append it to the given list of measurements
        Logging will indicate it is the result of a query
        """
        cls._log_measurement(measurement)
        measurements.append(measurement)

    @classmethod
    def _log_and_append_derived_measurements(
            cls,
            measurements: List[Measurement],
            derived_measurements: List[Measurement]):
        """
        Convenience method to log a list of derived measurements and append them to the given list of measurements.
        Logging will indicate it is a derived measurement
        """
        for derived_measurement in derived_measurements:
            cls._log_measurement(derived_measurement, is_derived=True)
            measurements.append(derived_measurement)

    @classmethod
    def _log_measurement(cls, measurement, is_derived: bool = False):
        measurement_type = "Derived" if is_derived else "Query"
        logging.debug(f'{measurement_type} measurement: {measurement}')

    def _flush_measurements(self, measurements: List[Measurement]):
        """
        Adds the measurements to the scan result and sends the measurements to the Soda Server if that's configured
        """
        self.scan_result.measurements.extend(measurements)
        if self.soda_server_client and measurements:
            measurement_jsons = [measurement.to_json() for measurement in measurements]
            self.soda_server_client.scan_measurements(self.scan_reference, measurement_jsons)

    def _flush_test_results(self, test_results: List[TestResult]):
        """
        Adds the test_results to the scan result and sends the measurements to the Soda Server if that's configured
        """
        self.scan_result.test_results.extend(test_results)
        if self.soda_server_client and test_results:
            test_result_jsons = [test_result.to_json() for test_result in test_results]
            self.soda_server_client.scan_test_results(self.scan_reference, test_result_jsons)

    def _ensure_scan_reference(self):
        if self.soda_server_client and not self.scan_reference:
            self.start_scan_response = self.soda_server_client.scan_start(
                self.warehouse.name,
                self.warehouse.dialect.type,
                self.scan_yml.table_name,
                self.time)
            self.scan_reference = self.start_scan_response['scanReference']



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
from typing import List

from sodasql.scan.db import sql_fetchone, sql_fetchall, sql_fetchone_description, sql_fetchall_description
from sodasql.scan.dialect import Dialect
from sodasql.scan.warehouse_yml import WarehouseYml


class Warehouse:

    def __init__(self, warehouse_yml: WarehouseYml):
        self.name = warehouse_yml.name
        self.dialect: Dialect = warehouse_yml.dialect
        self.connection = self.dialect.create_connection()

    def sql_fetchone(self, sql) -> tuple:
        return sql_fetchone(self.connection, sql)

    def sql_fetchone_description(self, sql) -> tuple:
        return sql_fetchone_description(self.connection, sql)

    def sql_fetchall(self, sql) -> List[tuple]:
        return sql_fetchall(self.connection, sql)

    def sql_fetchall_description(self, sql) -> tuple:
        return sql_fetchall_description(self.connection, sql)

    def create_scan(self, *args, **kwargs):
        return self.dialect.create_scan(self, *args, **kwargs)

    def close(self):
        if self.connection:
            try:
                self.connection.close()
            except Exception as e:
                logging.debug(f'Closing connection failed: {str(e)}')




#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import psycopg2

from sodasql.scan.dialect import Dialect, POSTGRES, KEY_WAREHOUSE_TYPE
from sodasql.scan.parser import Parser


class PostgresDialect(Dialect):

    def __init__(self, parser: Parser, type: str = POSTGRES):
        super().__init__(type)
        self.host = parser.get_str_optional_env('host', 'localhost')
        self.port = parser.get_str_optional_env('port', '5432')
        self.username = parser.get_str_required_env('username')
        self.password = parser.get_credential('password')
        self.database = parser.get_str_required_env('database')
        self.schema = parser.get_str_required_env('schema')

    def default_connection_properties(self, params: dict):
        return {
            KEY_WAREHOUSE_TYPE: POSTGRES,
            'host': 'localhost',
            'username': 'env_var(POSTGRES_USERNAME)',
            'password': 'env_var(POSTGRES_PASSWORD)',
            'database': params.get('database', 'your_database'),
            'schema': 'public'
        }

    def default_env_vars(self, params: dict):
        return {
            'POSTGRES_USERNAME': params.get('username', 'Eg johndoe'),
            'POSTGRES_PASSWORD': params.get('password', 'Eg abc123')
        }

    def sql_tables_metadata_query(self, limit: str = 10, filter: str = None):
        return (f"SELECT table_name \n" 
                f"FROM information_schema.tables \n" 
                f"WHERE lower(table_schema)='{self.schema.lower()}'")

    def sql_connection_test(self):
        pass

    def create_connection(self):
        return psycopg2.connect(
            user=self.username,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.database,
            options=f'-c search_path={self.schema}' if self.schema else None)

    def sql_columns_metadata_query(self, table_name: str) -> str:
        sql = (f"SELECT column_name, data_type, is_nullable \n"
               f"FROM information_schema.columns \n"
               f"WHERE lower(table_name) = '{table_name}'")
        if self.database:
            sql += f" \n  AND table_catalog = '{self.database}'"
        if self.schema:
            sql += f" \n  AND table_schema = '{self.schema}'"
        return sql

    def qualify_table_name(self, table_name: str) -> str:
        if self.schema:
            return f'"{self.schema}"."{table_name}"'
        return f'"{table_name}"'

    def sql_expr_regexp_like(self, expr: str, pattern: str):
        return f"{expr} ~* '{self.qualify_regex(pattern)}'"

    def sql_expr_cast_text_to_number(self, quoted_column_name, validity_format):
        if validity_format == 'number_whole':
            return f"CAST({quoted_column_name} AS {self.data_type_decimal})"
        not_number_pattern = self.qualify_regex(r"[^-\d\.\,]")
        comma_pattern = self.qualify_regex(r"\,")
        return f"CAST(REGEXP_REPLACE(REGEXP_REPLACE({quoted_column_name}, '{not_number_pattern}', '', 'g'), "\
               f"'{comma_pattern}', '.', 'g') AS {self.data_type_decimal})"

    def get_type_name(self, column_description):
        return PostgresDialect.type_names_by_type_code.get(str(column_description[1]))

    type_names_by_type_code = {
        '16': 'bool',
        '17': 'bytea',
        '18': 'char',
        '19': 'name',
        '20': 'int8',
        '21': 'int2',
        '22': 'int2vector',
        '23': 'int4',
        '24': 'regproc',
        '25': 'text',
        '26': 'oid',
        '27': 'tid',
        '28': 'xid',
        '29': 'cid',
        '30': 'oidvector',
        '32': 'pg_ddl_command',
        '71': 'pg_type',
        '75': 'pg_attribute',
        '81': 'pg_proc',
        '83': 'pg_class',
        '114': 'json',
        '142': 'xml',
        '143': '_xml',
        '194': 'pg_node_tree',
        '199': '_json',
        '210': 'smgr',
        '325': 'index_am_handler',
        '600': 'point',
        '601': 'lseg',
        '602': 'path',
        '603': 'box',
        '604': 'polygon',
        '628': 'line',
        '629': '_line',
        '650': 'cidr',
        '651': '_cidr',
        '700': 'float4',
        '701': 'float8',
        '702': 'abstime',
        '703': 'reltime',
        '704': 'tinterval',
        '705': 'unknown',
        '718': 'circle',
        '719': '_circle',
        '790': 'money',
        '791': '_money',
        '829': 'macaddr',
        '869': 'inet',
        '1000': '_bool',
        '1001': '_bytea',
        '1002': '_char',
        '1003': '_name',
        '1005': '_int2',
        '1006': '_int2vector',
        '1007': '_int4',
        '1008': '_regproc',
        '1009': '_text',
        '1010': '_tid',
        '1011': '_xid',
        '1012': '_cid',
        '1013': '_oidvector',
        '1014': '_bpchar',
        '1015': '_varchar',
        '1016': '_int8',
        '1017': '_point',
        '1018': '_lseg',
        '1019': '_path',
        '1020': '_box',
        '1021': '_float4',
        '1022': '_float8',
        '1023': '_abstime',
        '1024': '_reltime',
        '1025': '_tinterval',
        '1027': '_polygon',
        '1028': '_oid',
        '1033': 'aclitem',
        '1034': '_aclitem',
        '1040': '_macaddr',
        '1041': '_inet',
        '1042': 'bpchar',
        '1043': 'varchar',
        '1082': 'date',
        '1083': 'time',
        '1114': 'timestamp',
        '1115': '_timestamp',
        '1182': '_date',
        '1183': '_time',
        '1184': 'timestamptz',
        '1185': '_timestamptz',
        '1186': 'interval',
        '1187': '_interval',
        '1231': '_numeric',
        '1248': 'pg_database',
        '1263': '_cstring',
        '1266': 'timetz',
        '1270': '_timetz',
        '1560': 'bit',
        '1561': '_bit',
        '1562': 'varbit',
        '1563': '_varbit',
        '1700': 'numeric',
        '1790': 'refcursor',
        '2201': '_refcursor',
        '2202': 'regprocedure',
        '2203': 'regoper',
        '2204': 'regoperator',
        '2205': 'regclass',
        '2206': 'regtype',
        '2207': '_regprocedure',
        '2208': '_regoper',
        '2209': '_regoperator',
        '2210': '_regclass',
        '2211': '_regtype',
        '2249': 'record',
        '2275': 'cstring',
        '2276': 'any',
        '2277': 'anyarray',
        '2278': 'void',
        '2279': 'trigger',
        '2280': 'language_handler',
        '2281': 'internal',
        '2282': 'opaque',
        '2283': 'anyelement',
        '2287': '_record',
        '2776': 'anynonarray',
        '2842': 'pg_authid',
        '2843': 'pg_auth_members',
        '2949': '_txid_snapshot',
        '2950': 'uuid',
        '2951': '_uuid',
        '2970': 'txid_snapshot',
        '3115': 'fdw_handler',
        '3220': 'pg_lsn',
        '3221': '_pg_lsn',
        '3310': 'tsm_handler',
        '3500': 'anyenum',
        '3614': 'tsvector',
        '3615': 'tsquery',
        '3642': 'gtsvector',
        '3643': '_tsvector',
        '3644': '_gtsvector',
        '3645': '_tsquery',
        '3734': 'regconfig',
        '3735': '_regconfig',
        '3769': 'regdictionary',
        '3770': '_regdictionary',
        '3802': 'jsonb',
        '3807': '_jsonb',
        '3831': 'anyrange',
        '3838': 'event_trigger',
        '3904': 'int4range',
        '3905': '_int4range',
        '3906': 'numrange',
        '3907': '_numrange',
        '3908': 'tsrange',
        '3909': '_tsrange',
        '3910': 'tstzrange',
        '3911': '_tstzrange',
        '3912': 'daterange',
        '3913': '_daterange',
        '3926': 'int8range',
        '3927': '_int8range',
        '4066': 'pg_shseclabel',
        '4089': 'regnamespace',
        '4090': '_regnamespace',
        '4096': 'regrole',
        '4097': '_regrole',
        '10000': 'pg_attrdef',
        '10001': 'pg_constraint',
        '10002': 'pg_inherits',
        '10003': 'pg_index',
        '10004': 'pg_operator',
        '10005': 'pg_opfamily',
        '10006': 'pg_opclass',
        '10156': 'pg_am',
        '10157': 'pg_amop',
        '10846': 'pg_amproc',
        '11393': 'pg_language',
        '11394': 'pg_largeobject_metadata',
        '11395': 'pg_largeobject',
        '11396': 'pg_aggregate',
        '11397': 'pg_statistic',
        '11398': 'pg_rewrite',
        '11399': 'pg_trigger',
        '11400': 'pg_event_trigger',
        '11401': 'pg_description',
        '11402': 'pg_cast',
        '11616': 'pg_enum',
        '11617': 'pg_namespace',
        '11618': 'pg_conversion',
        '11619': 'pg_depend',
        '11620': 'pg_db_role_setting',
        '11621': 'pg_tablespace',
        '11622': 'pg_pltemplate',
        '11623': 'pg_shdepend',
        '11624': 'pg_shdescription',
        '11625': 'pg_ts_config',
        '11626': 'pg_ts_config_map',
        '11627': 'pg_ts_dict',
        '11628': 'pg_ts_parser',
        '11629': 'pg_ts_template',
        '11630': 'pg_extension',
        '11631': 'pg_foreign_data_wrapper',
        '11632': 'pg_foreign_server',
        '11633': 'pg_user_mapping',
        '11634': 'pg_foreign_table',
        '11635': 'pg_policy',
        '11636': 'pg_replication_origin',
        '11637': 'pg_default_acl',
        '11638': 'pg_init_privs',
        '11639': 'pg_seclabel',
        '11640': 'pg_collation',
        '11641': 'pg_range',
        '11642': 'pg_transform',
        '11643': 'pg_toast_2604',
        '11644': 'pg_toast_2606',
        '11645': 'pg_toast_2609',
        '11646': 'pg_toast_1255',
        '11647': 'pg_toast_2618',
        '11648': 'pg_toast_3596',
        '11649': 'pg_toast_2619',
        '11650': 'pg_toast_2620',
        '11651': 'pg_toast_2396',
        '11652': 'pg_toast_2964',
        '11653': 'pg_toast_3592',
        '11655': 'pg_roles',
        '11659': 'pg_shadow',
        '11662': 'pg_group',
        '11665': 'pg_user',
        '11668': 'pg_policies',
        '11672': 'pg_rules',
        '11676': 'pg_views',
        '11680': 'pg_tables',
        '11684': 'pg_matviews',
        '11688': 'pg_indexes',
        '11692': 'pg_stats',
        '11696': 'pg_locks',
        '11699': 'pg_cursors',
        '11702': 'pg_available_extensions',
        '11705': 'pg_available_extension_versions',
        '11708': 'pg_prepared_xacts',
        '11712': 'pg_prepared_statements',
        '11715': 'pg_seclabels',
        '11719': 'pg_settings',
        '11724': 'pg_file_settings',
        '11727': 'pg_timezone_abbrevs',
        '11730': 'pg_timezone_names',
        '11733': 'pg_config',
        '11736': 'pg_stat_all_tables',
        '11740': 'pg_stat_xact_all_tables',
        '11744': 'pg_stat_sys_tables',
        '11748': 'pg_stat_xact_sys_tables',
        '11751': 'pg_stat_user_tables',
        '11755': 'pg_stat_xact_user_tables',
        '11758': 'pg_statio_all_tables',
        '11762': 'pg_statio_sys_tables',
        '11765': 'pg_statio_user_tables',
        '11768': 'pg_stat_all_indexes',
        '11772': 'pg_stat_sys_indexes',
        '11775': 'pg_stat_user_indexes',
        '11778': 'pg_statio_all_indexes',
        '11782': 'pg_statio_sys_indexes',
        '11785': 'pg_statio_user_indexes',
        '11788': 'pg_statio_all_sequences',
        '11792': 'pg_statio_sys_sequences',
        '11795': 'pg_statio_user_sequences',
        '11798': 'pg_stat_activity',
        '11802': 'pg_stat_replication',
        '11806': 'pg_stat_wal_receiver',
        '11809': 'pg_stat_ssl',
        '11812': 'pg_replication_slots',
        '11816': 'pg_stat_database',
        '11819': 'pg_stat_database_conflicts',
        '11822': 'pg_stat_user_functions',
        '11826': 'pg_stat_xact_user_functions',
        '11830': 'pg_stat_archiver',
        '11833': 'pg_stat_bgwriter',
        '11836': 'pg_stat_progress_vacuum',
        '11840': 'pg_user_mappings',
        '11844': 'pg_replication_origin_status',
        '12129': 'cardinal_number',
        '12131': 'character_data',
        '12132': 'sql_identifier',
        '12134': 'information_schema_catalog_name',
        '12136': 'time_stamp',
        '12137': 'yes_or_no',
        '12140': 'applicable_roles',
        '12144': 'administrable_role_authorizations',
        '12147': 'attributes',
        '12151': 'character_sets',
        '12155': 'check_constraint_routine_usage',
        '12159': 'check_constraints',
        '12163': 'collations',
        '12166': 'collation_character_set_applicability',
        '12169': 'column_domain_usage',
        '12173': 'column_privileges',
        '12177': 'column_udt_usage',
        '12181': 'columns',
        '12185': 'constraint_column_usage',
        '12189': 'constraint_table_usage',
        '12193': 'domain_constraints',
        '12197': 'domain_udt_usage',
        '12200': 'domains',
        '12204': 'enabled_roles',
        '12207': 'key_column_usage',
        '12211': 'parameters',
        '12215': 'referential_constraints',
        '12219': 'role_column_grants',
        '12222': 'routine_privileges',
        '12226': 'role_routine_grants',
        '12229': 'routines',
        '12233': 'schemata',
        '12236': 'sequences',
        '12240': 'sql_features',
        '12242': 'pg_toast_12239',
        '12245': 'sql_implementation_info',
        '12247': 'pg_toast_12244',
        '12250': 'sql_languages',
        '12252': 'pg_toast_12249',
        '12255': 'sql_packages',
        '12257': 'pg_toast_12254',
        '12260': 'sql_parts',
        '12262': 'pg_toast_12259',
        '12265': 'sql_sizing',
        '12267': 'pg_toast_12264',
        '12270': 'sql_sizing_profiles',
        '12272': 'pg_toast_12269',
        '12275': 'table_constraints',
        '12279': 'table_privileges',
        '12283': 'role_table_grants',
        '12286': 'tables',
        '12290': 'transforms',
        '12294': 'triggered_update_columns',
        '12298': 'triggers',
        '12302': 'udt_privileges',
        '12306': 'role_udt_grants',
        '12309': 'usage_privileges',
        '12313': 'role_usage_grants',
        '12316': 'user_defined_types',
        '12320': 'view_column_usage',
        '12324': 'view_routine_usage',
        '12328': 'view_table_usage',
        '12332': 'views',
        '12336': 'data_type_privileges',
        '12340': 'element_types',
        '12344': '_pg_foreign_table_columns',
        '12348': 'column_options',
        '12351': '_pg_foreign_data_wrappers',
        '12354': 'foreign_data_wrapper_options',
        '12357': 'foreign_data_wrappers',
        '12360': '_pg_foreign_servers',
        '12364': 'foreign_server_options',
        '12367': 'foreign_servers',
        '12370': '_pg_foreign_tables',
        '12374': 'foreign_table_options',
        '12377': 'foreign_tables',
        '12380': '_pg_user_mappings',
        '12384': 'user_mapping_options',
        '12388': 'user_mappings',
        '24626': '_level2emissions_20200529',
        '24627': 'level2emissions_20200529',
        '24629': 'pg_toast_24625',
        '24632': '_level2emissions_20200630',
        '24633': 'level2emissions_20200630',
        '24635': 'pg_toast_24631',
        '24672': '_may',
        '24673': 'may',
        '24675': 'pg_toast_24671',
        '24678': '_june',
        '24679': 'june',
        '24681': 'pg_toast_24677',
    }



import pathlib
import re
from os import path

from setuptools import setup, find_namespace_packages

readme = (pathlib.Path(__file__).parent / "README.md").read_text()

with open("requirements.txt", "r", encoding="utf-8") as file:
    requires = [line.strip() for line in file if line.strip()]


def get_version():
    with open(path.join("sodasql", "version.py")) as f:
        contents = f.read()
        matches = re.findall(r"SODA_SQL_VERSION = '([^']+)'", contents)
        return matches[0] if len(matches) > 0 else '0.0.0'


setup(
    name="soda-sql",
    version=get_version(),
    author="Tom Baeyens",
    author_email="tom@soda.io",
    description="Soda SQL library & CLI",
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=find_namespace_packages(include=["sodasql*"]),
    install_requires=requires,
    entry_points={"console_scripts": ["soda=sodasql.cli.cli:main"]},
)



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from dataclasses import dataclass
from typing import Set, List

from sodasql.scan.missing import Missing
from sodasql.scan.sql_metric_yml import SqlMetricYml
from sodasql.scan.test import Test
from sodasql.scan.validity import Validity


@dataclass
class ScanYmlColumn:

    metrics: Set[str]
    sql_metric_ymls: List[SqlMetricYml]
    missing: Missing
    validity: Validity
    tests: List[Test]



#  Copyright 2020 Soda
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#   http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
from json.decoder import JSONDecodeError
import re

from google.cloud import bigquery
from google.cloud.bigquery import dbapi
from google.oauth2.service_account import Credentials

from sodasql.scan.dialect import Dialect, BIGQUERY, KEY_WAREHOUSE_TYPE
from sodasql.scan.parser import Parser


class BigQueryDialect(Dialect):

    data_type_varchar_255 = "STRING"
    data_type_integer = "INT64"
    data_type_decimal = "DECIMAL"
    data_type_bigint = "BIGNUMERIC"

    def __init__(self, parser: Parser):
        super().__init__(BIGQUERY)
        self.account_info_dict = self.__parse_json_credential('account_info_json', parser)
        self.dataset_name = parser.get_str_required('dataset')
        self.client = None

    def default_connection_properties(self, params: dict):
        return {
            KEY_WAREHOUSE_TYPE: BIGQUERY,
            'account_info': 'env_var(BIGQUERY_ACCOUNT_INFO)',
            'dataset': params.get('database', 'Eg your_bigquery_dataset')
        }

    def default_env_vars(self, params: dict):
        return {
            'BIGQUERY_ACCOUNT_INFO': '...'
        }

    def create_connection(self):
        credentials = Credentials.from_service_account_info(self.account_info_dict)
        project_id = self.account_info_dict['project_id']
        self.client = bigquery.Client(project=project_id, credentials=credentials)
        return dbapi.Connection(self.client)

    def sql_tables_metadata_query(self, limit: str = 10, filter: str = None):
        return (f"SELECT table_name \n"
                f"FROM `{self.dataset_name}.INFORMATION_SCHEMA.TABLES`;")

    def sql_columns_metadata_query(self, table_name: str):
        return (f"SELECT column_name, data_type, is_nullable "
                f'FROM `{self.dataset_name}.INFORMATION_SCHEMA.COLUMNS` '
                f"WHERE table_name = '{table_name}';")

    def qualify_table_name(self, table_name: str) -> str:
        return f'`{self.dataset_name}.{table_name}`'

    def qualify_writable_table_name(self, table_name: str) -> str:
        return self.qualify_table_name(table_name)

    def sql_expr_regexp_like(self, expr: str, pattern: str):
        return f"REGEXP_CONTAINS({expr}, r'{self.qualify_regex(pattern)}')"

    def qualify_table_name(self, table_name: str) -> str:
        return f'`{self.dataset_name}.{table_name}`'

    def qualify_writable_table_name(self, table_name: str) -> str:
        return self.qualify_table_name(table_name)

    def qualify_regex(self, regex):
        return regex.replace("''", "\\'")

    def qualify_string(self, value: str):
        return self.qualify_regex(value)

    def sql_expr_regexp_like(self, expr: str, pattern: str):
        return f"REGEXP_CONTAINS({expr}, r'{self.qualify_regex(pattern)}')"

    @staticmethod
    def __parse_json_credential(credential_name, parser):
        try:
            return json.loads(parser.get_credential(credential_name))
        except JSONDecodeError as e:
            parser.error(f'Error parsing credential %s: %s', credential_name, e)

    def sql_expr_cast_text_to_number(self, quoted_column_name, validity_format):
        if validity_format == 'number_whole':
            return f"CAST({quoted_column_name} AS {self.data_type_decimal})"
        not_number_pattern = self.qualify_regex(r"[^-\d\.\,]")
        comma_pattern = self.qualify_regex(r"\,")
        return f"CAST(REGEXP_REPLACE(REGEXP_REPLACE({quoted_column_name}, r'{not_number_pattern}', ''), "\
               f"r'{comma_pattern}', '.') AS {self.data_type_decimal})"
